import os

from .. import PluginBase


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['paths']

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': ['openpaperwork_core.app'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python']
            },
        ]

    def paths_get_config_dir(self):
        config_dir = self.core.call_success(
            "fs_safe",
            os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
        )
        self.core.call_success("fs_mkdir_p", config_dir)

        if os.name == 'nt':  # hide ~/.local on Windows
            self.core.call_success("fs_hide", config_dir)

        return config_dir

    def paths_get_data_dir(self):
        local_dir = os.path.expanduser("~/.local")
        data_dir = os.getenv("XDG_DATA_HOME", os.path.join(local_dir, "share"))

        data_dir = self.core.call_success("fs_safe", data_dir)

        app_name = self.core.call_success("app_get_fs_name")
        app_dir = self.core.call_success("fs_join", data_dir, app_name)
        self.core.call_success("fs_mkdir_p", app_dir)

        if os.name == 'nt':  # hide ~/.local on Windows
            local_dir = self.core.call_success("fs_safe", local_dir)
            if self.core.call_success("fs_isdir", local_dir) is not None:
                self.core.call_success("fs_hide", local_dir)

        return app_dir



import openpaperwork_core

from ... import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'shortcuts',
            'shortcuts_app',
            'shortcuts_app_find',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_app_find',
                'defaults': ['paperwork_gtk.actions.app.find'],
            },
            {
                'interface': 'app_shortcuts',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
        ]

    def init(self, core):
        super().init(core)

    def on_gtk_initialized(self):
        self.core.call_all(
            "app_shortcut_add",
            _("Global"), _("Find"),
            "<Control>f", "win.app_find"
        )

        # TODO
        # self.core.call_all(
        #     "app_shortcut_add",
        #     _("Global"), _("Find the next match"),
        #     "<Control>G", "win.app_find_next"
        # )
        # self.core.call_all(
        #     "app_shortcut_add",
        #     _("Global"), _("Find the previous match"),
        #     "<Control><Shift>G", "win.app_find_prev"
        # )



import ctypes
import io
import logging
import os
import tempfile

try:
    from gi.repository import Gio
    from gi.repository import GLib
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core.deps
import openpaperwork_core.fs


LOGGER = logging.getLogger(__name__)


class _GioFileAdapter(io.RawIOBase):
    def __init__(self, gfile, mode='r'):
        super().__init__()
        self.gfile = gfile
        self.mode = mode

        if 'w' in mode and 'r' not in mode:
            self.size = 0
        elif ('w' in mode or 'a' in mode) and not gfile.query_exists():
            self.size = 0
        else:
            try:
                fi = gfile.query_info(
                    Gio.FILE_ATTRIBUTE_STANDARD_SIZE,
                    Gio.FileQueryInfoFlags.NONE
                )
                self.size = fi.get_attribute_uint64(
                    Gio.FILE_ATTRIBUTE_STANDARD_SIZE
                )
            except GLib.GError as exc:
                LOGGER.warning("Gio.Gerror", exc_info=exc)
                raise IOError(str(exc))

        self.gfd = None
        self.gin = None
        self.gout = None

        if 'r' in mode and 'w' in mode:
            if gfile.query_exists():
                self.gfd = gfile.open_readwrite()
            else:
                # create_readwrite() doesn't seem to always work on
                # Windows+MSYS2
                self.gfd = gfile.create_readwrite(Gio.FileCreateFlags.PRIVATE)
            if 'w' in mode:
                self.gfd.seek(0, GLib.SeekType.SET)
                self.gfd.truncate(0)
            self.gin = self.gfd.get_input_stream()
            self.gout = self.gfd.get_output_stream()
        elif 'r' in mode:
            self.gfd = gfile.read()
            self.gin = self.gfd
        elif 'w' in mode or 'a' in mode:
            if 'w' in mode:
                self.gfd = gfile.replace(
                    None,  # etag
                    False,  # make_backup
                    Gio.FileCreateFlags.PRIVATE
                )
            elif 'a' in mode:
                self.gfd = gfile.append_to(
                    Gio.FileCreateFlags.PRIVATE
                )
            self.gout = self.gfd

    def readable(self):
        return True

    def writable(self):
        return 'w' in self.mode or 'a' in self.mode

    def read(self, size=-1):
        if not self.readable():
            raise OSError("File is not readable")
        if size <= 0:
            size = self.size
            if size <= 0:
                return b""
        assert size > 0
        return self.gin.read_bytes(size).get_data()

    def readall(self):
        return self.read(-1)

    def readinto(self, b):
        raise OSError("readinto() not supported on Gio.File objects")

    def readline(self, size=-1):
        raise OSError("readline() not supported on Gio.File objects")

    def readlines(self, hint=-1):
        LOGGER.warning("readlines() shouldn't be called on a binary file"
                       " descriptor. This is not cross-platform")
        return [(x + b"\n") for x in self.readall().split(b"\n")]

    def seek(self, offset, whence=os.SEEK_SET):
        whence = {
            os.SEEK_CUR: GLib.SeekType.CUR,
            os.SEEK_END: GLib.SeekType.END,
            os.SEEK_SET: GLib.SeekType.SET,
        }[whence]
        self.gfd.seek(offset, whence)

    def seekable(self):
        return True

    def tell(self):
        return self.gin.tell()

    def flush(self):
        pass

    def truncate(self, size=None):
        if size is None:
            size = self.tell()
        self.gfd.truncate(size)

    def fileno(self):
        if self.gout is not None and hasattr(self.gout, 'get_fd'):
            return self.gout.get_fd()
        if self.gin is not None and hasattr(self.gin, 'get_fd'):
            return self.gin.get_fd()
        raise io.UnsupportedOperation("fileno() on Gio.File unsupported")

    def isatty(self):
        return False

    def write(self, b):
        res = self.gout.write_all(b)
        if not res[0]:
            raise OSError("write_all() failed on {}: {}".format(
                self.gfile.get_uri(), res)
            )
        return res[1]

    def writelines(self, lines):
        self.write(b"".join(lines))

    def close(self):
        self.flush()
        super().close()
        if self.gin is not None and self.gin is not self.gfd:
            self.gin.close()
        if self.gout is not None and self.gout is not self.gfd:
            self.gout.close()
        if self.gfd:
            self.gfd.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class _GioUTF8FileAdapter(io.RawIOBase):
    def __init__(self, raw):
        super().__init__()
        self.raw = raw
        self.line_iterator = None

    def readable(self):
        return self.raw.readable()

    def writable(self):
        return self.raw.writable()

    def read(self, *args, **kwargs):
        r = self.raw.read(*args, **kwargs)
        return r.decode("utf-8")

    def readall(self, *args, **kwargs):
        r = self.raw.readall(*args, **kwargs)
        return r.decode("utf-8")

    def readinto(self, *args, **kwargs):
        r = self.raw.readinto(*args, **kwargs)
        return r.decode("utf-8")

    def readlines(self, hint=-1):
        all = self.readall()
        if os.linesep != "\n":
            all = all.replace(os.linesep, "\n")
        lines = [(x + "\n") for x in all.split("\n")]
        if lines[-1] == "\n":
            return lines[:-1]
        return lines

    def readline(self, hint=-1):
        if self.line_iterator is None:
            self.line_iterator = (line for line in self.readlines(hint))
        try:
            return next(self.line_iterator)
        except StopIteration:
            return ''

    def seek(self, *args, **kwargs):
        return self.raw.seek(*args, **kwargs)

    def seekable(self, seekable):
        return self.raw.seekable()

    @property
    def closed(self):
        return self.raw.closed

    def tell(self):
        # XXX(Jflesch): wrong ...
        return self.raw.tell()

    def flush(self):
        return self.raw.flush()

    def truncate(self, *args, **kwargs):
        # XXX(Jflesch): wrong ...
        return self.raw.truncate(*args, **kwargs)

    def fileno(self):
        return self.raw.fileno()

    def isatty(self):
        return self.raw.isatty()

    def write(self, b):
        b = b.encode("utf-8")
        return self.raw.write(b)

    def writelines(self, lines):
        lines = [
            (line + os.linesep).encode("utf-8")
            for line in lines
        ]
        return self.raw.writelines(lines)

    def close(self):
        self.raw.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class Plugin(openpaperwork_core.fs.CommonFsPluginBase):
    PRIORITY = 50

    def __init__(self):
        super().__init__()
        self.vfs = None
        if GLIB_AVAILABLE:
            self.vfs = Gio.Vfs.get_default()
        self.tmp_files = set()

    def get_interfaces(self):
        return super().get_interfaces() + ['chkdeps']

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def fs_open(self, uri, mode='r', needs_fileno=False, **kwargs):
        if needs_fileno:
            # On Windows, `Gio.[Unix]OutputStream.get_fd()` doesn't seem
            # to be available
            return

        f = self.vfs.get_file_for_uri(uri)
        if ('w' not in mode and 'a' not in mode):
            if self.fs_exists(uri) is None:
                return None
        try:
            raw = _GioFileAdapter(f, mode)
            if 'b' in mode:
                return raw
            return _GioUTF8FileAdapter(raw)
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError("fs_open({}, mode={}): {}".format(
                uri, mode, str(exc)
            ))

    def fs_exists(self, url):
        if not GLIB_AVAILABLE:
            return None

        try:
            f = self.vfs.get_file_for_uri(url)
            if not f.query_exists():
                # this file does not exist for us, but it does not mean
                # another implementation of the plugin interface 'fs'
                # cannot handle it
                return None
            return True
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_listdir(self, url):
        if not GLIB_AVAILABLE:
            return None

        try:
            f = self.vfs.get_file_for_uri(url)
            if not f.query_exists():
                return None
            children = f.enumerate_children(
                Gio.FILE_ATTRIBUTE_STANDARD_NAME, Gio.FileQueryInfoFlags.NONE,
                None
            )
            for child in children:
                child = f.get_child(child.get_name())
                yield child.get_uri()
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_rename(self, old_url, new_url):
        try:
            old = self.vfs.get_file_for_uri(old_url)
            new = self.vfs.get_file_for_uri(new_url)
            assert not old.equal(new)
            if not old.query_exists():
                return None
            old.move(new, Gio.FileCopyFlags.NONE)
            return True
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_unlink(self, url, trash=True, **kwargs):
        try:
            f = self.vfs.get_file_for_uri(url)
            if not f.query_exists():
                return None

            LOGGER.info("Deleting %s (trash=%s) ...", url, trash)

            if not trash:
                deleted = f.delete()
                if not deleted:
                    raise IOError("Failed to delete %s" % url)
                return None

            deleted = False
            try:
                deleted = f.trash()
                if not deleted:
                    LOGGER.warning(
                        "Failed to trash %s. Will try to delete it instead",
                        url
                    )
            except Exception as exc:
                LOGGER.warning("Failed to trash %s. Will try to delete it"
                               " instead", url, exc_info=exc)

            if deleted and self.fs_exists(url) is not None:
                deleted = False
                # WORKAROUND(Jflesch): It seems in Flatpak, f.trash()
                # returns True when it actually does nothing.
                LOGGER.warning(
                    "trash(%s) returned True but file wasn't trashed."
                    " Will try to deleting it instead", f.get_uri()
                )

            if not deleted:
                try:
                    deleted = f.delete()
                except Exception as exc:
                    LOGGER.warning("Failed to deleted %s", url, exc_info=exc)
            if not deleted:
                raise IOError("Failed to delete %s" % url)
            return True
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_rm_rf(self, url, trash=True, **kwargs):
        if self.fs_exists(url) is None:
            return None

        try:
            LOGGER.info("Deleting %s ...", url)
            f = self.vfs.get_file_for_uri(url)
            deleted = False
            if trash:
                try:
                    deleted = f.trash()
                    if not deleted:
                        LOGGER.warning(
                            "Failed to trash %s."
                            " Will try to delete it instead",
                            url
                        )
                except Exception as exc:
                    LOGGER.warning(
                        "Failed to trash %s (trash()=%s)."
                        " Will try to delete it instead",
                        url, trash, exc_info=exc
                    )

            if deleted and self.fs_exists(url) is not None:
                deleted = False
                # WORKAROUND(Jflesch): It seems in Flatpak, f.trash()
                # returns True when it actually does nothing.
                LOGGER.warning(
                    "trash(%s) returned True but file wasn't trashed."
                    " Will try to deleting it instead", url
                )

            if not deleted:
                self._rm_rf(f)
            LOGGER.info("%s deleted", url)
            return True
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def _rm_rf(self, gfile):
        try:
            to_delete = [
                f for f in self._recurse(
                    gfile, dir_included=True, follow_symlinks=False
                )
            ]
            # make sure to delete the parent directory last:
            to_delete.sort(reverse=True, key=lambda f: f.get_uri())
            for f in to_delete:
                if not f.delete():
                    raise IOError("Failed to delete %s" % f.get_uri())
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_get_mtime(self, url):
        try:
            f = self.vfs.get_file_for_uri(url)
            if not f.query_exists():
                raise IOError("File {} does not exist".format(str(url)))
            if os.name == "nt":
                # WORKAROUND(Jflesch):
                # On Windows+MSYS2, it seems Gio.File.query_info()
                # return always 0 for Gio.FILE_ATTRIBUTE_TIME_CHANGED.
                path = self.fs_unsafe(url)
                return int(os.stat(path).st_mtime)
            fi = f.query_info(
                Gio.FILE_ATTRIBUTE_TIME_CHANGED, Gio.FileQueryInfoFlags.NONE
            )
            r = fi.get_attribute_uint64(Gio.FILE_ATTRIBUTE_TIME_CHANGED)
            if int(r) != 0:
                return r
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)

    def fs_getsize(self, url):
        try:
            f = self.vfs.get_file_for_uri(url)
            fi = f.query_info(
                Gio.FILE_ATTRIBUTE_STANDARD_SIZE, Gio.FileQueryInfoFlags.NONE
            )
            return fi.get_attribute_uint64(Gio.FILE_ATTRIBUTE_STANDARD_SIZE)
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_isdir(self, url):
        if not GLIB_AVAILABLE:
            return None

        try:
            f = self.vfs.get_file_for_uri(url)
            if not f.query_exists():
                return None
            fi = f.query_info(
                Gio.FILE_ATTRIBUTE_STANDARD_TYPE, Gio.FileQueryInfoFlags.NONE
            )
            if fi.get_file_type() == Gio.FileType.DIRECTORY:
                return True
            return None
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def fs_copy(self, old_url, new_url):
        try:
            old = self.vfs.get_file_for_uri(old_url)
            new = self.vfs.get_file_for_uri(new_url)
            if new.query_exists():
                new.delete()
            old.copy(new, Gio.FileCopyFlags.ALL_METADATA)
            return new_url
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            return None

    def fs_mkdir_p(self, url):
        if not GLIB_AVAILABLE:
            return None

        try:
            f = self.vfs.get_file_for_uri(url)
            if not f.query_exists():
                if os.name == "nt":
                    # WORKAROUND(Jflesch): On Windows+MSYS2,
                    # Gio.File.make_directory_with_parents() raises
                    # a Gio.GError "Unsupported operation" (bug ?)
                    path = self.fs_unsafe(url)
                    os.makedirs(path, mode=0o700, exist_ok=True)
                    return True
                f.make_directory_with_parents()
        except GLib.GError as exc:
            LOGGER.warning("Gio.GError", exc_info=exc)
            raise IOError("fs_mkdir_p({}): {}".format(url, str(exc)))

        return True

    def _recurse(self, parent, dir_included=False, follow_symlinks=True):
        """
        Yield all the children (depth first), but not the parent.
        """
        try:
            children = parent.enumerate_children(
                (
                    Gio.FILE_ATTRIBUTE_STANDARD_NAME
                    + "," + Gio.FILE_ATTRIBUTE_STANDARD_IS_SYMLINK
                ),
                Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                None
            )
        except GLib.GError:
            # assumes it's a file and not a directory
            yield parent
            return

        for child in children:
            name = child.get_name()
            if not follow_symlinks:
                is_symlink = child.has_attribute("standard::is-symlink")
                is_symlink = is_symlink and child.get_is_symlink()
                if is_symlink:
                    yield parent.get_child(name)
                    continue
            child = parent.get_child(name)
            try:
                for sub in self._recurse(
                        child, dir_included=dir_included,
                        follow_symlinks=follow_symlinks):
                    yield sub
            except GLib.GError:
                yield child

        if dir_included:
            yield parent

    def fs_recurse(self, parent_uri, dir_included=False):
        parent = self.vfs.get_file_for_uri(parent_uri)
        for f in self._recurse(parent, dir_included):
            yield f.get_uri()

    def fs_hide(self, uri):
        if os.name != 'nt':
            LOGGER.warning("fs_hide('%s') can only works on Windows", uri)
            return None
        filepath = self.fs_unsafe(uri)
        LOGGER.info("Hiding file: {}".format(filepath))
        ret = ctypes.windll.kernel32.SetFileAttributesW(
            filepath, 0x02  # hidden
        )
        if not ret:
            raise ctypes.WinError()
        return True

    def fs_get_mime(self, uri):
        if os.name == 'nt':
            # WORKAROUND(Jflesch):
            # Gio.File.query_info().get_content_type() returns crap on Windows
            # (for instance '.pdf' instead of 'application/pdf').
            return None

        gfile = self.vfs.get_file_for_uri(uri)
        info = gfile.query_info(
            "standard::content-type", Gio.FileQueryInfoFlags.NONE
        )
        return info.get_content_type()

    def fs_mktemp(self, prefix=None, suffix=None, mode='w+b', **kwargs):
        if 'b' not in mode:
            tmp = tempfile.NamedTemporaryFile(
                prefix=prefix, suffix=suffix, delete=False, mode=mode,
                encoding='utf-8'
            )
        else:
            tmp = tempfile.NamedTemporaryFile(
                prefix=prefix, suffix=suffix, delete=False, mode=mode
            )
        self.tmp_files.add(tmp.name)
        return (self.fs_safe(tmp.name), tmp)

    def fs_iswritable(self, url):
        try:
            f = self.vfs.get_file_for_uri(url)
            fi = f.query_info(
                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE,
                Gio.FileQueryInfoFlags.NONE
            )
            return fi.get_attribute_boolean(
                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE
            )
        except GLib.GError as exc:
            LOGGER.warning("Gio.Gerror", exc_info=exc)
            raise IOError(str(exc))

    def on_quit(self):
        for tmp_file in self.tmp_files:
            try:
                os.unlink(tmp_file)
            except FileNotFoundError:
                pass



import logging

import openpaperwork_core

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_properties"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def __init__(self):
        super().__init__()
        self.active_doc = None

    def get_interfaces(self):
        return [
            'menu',
            'menu_doc',
            'menu_doc_properties',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_actions',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all(
            "add_doc_main_action",
            "document-properties-symbolic",
            _("Document properties"),
            self._open_properties
        )

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def doc_close(self):
        self.active_doc = None

    def _open_properties(self, *args, **kwargs):
        assert self.active_doc is not None
        active = self.active_doc

        LOGGER.info("Opening properties of document %s", active[0])
        self.core.call_all("open_doc_properties", *active)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_redo_ocr"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = (None, None)
        self.action = None

    def get_interfaces(self):
        return [
            'action',
            'action_doc',
            'action_doc_redo_ocr',
            'chkdeps',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'ocr',
                'defaults': ['paperwork_backend.guesswork.ocr.pyocr'],
            },
            {
                'interface': 'ocr_settings',
                'defaults': ['paperwork_backend.pyocr'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        self.action = Gio.SimpleAction.new(ACTION_NAME, None)
        self.action.connect("activate", self._redo_ocr)
        self.core.call_all("app_actions_add", self.action)
        self._update_sensitivity()
        self.core.call_all(
            "ocr_add_observer_on_enabled",
            self._update_sensitivity
        )

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def doc_close(self):
        self.active_doc = (None, None)

    def _update_sensitivity(self):
        enabled = self.core.call_success("ocr_is_enabled")
        self.action.set_enabled(enabled is not None)

    def _redo_ocr(self, action, parameter):
        assert self.active_doc is not None
        (doc_id, doc_url) = self.active_doc

        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            LOGGER.warning("No pages in document %s. Nothing to do", doc_id)
            return

        LOGGER.info(
            "Will redo OCR on document %s (nb pages = %d)", doc_id, nb_pages
        )

        promise = openpaperwork_core.promise.Promise(
            self.core, self.core.call_all,
            args=("on_progress", "redo_ocr", 0.0, _("OCR on %s") % doc_id)
        )
        promise = promise.then(lambda *args, **kwargs: None)

        for page_idx in range(0, nb_pages):
            promise = promise.then(
                self.core.call_all, "on_progress", "redo_ocr",
                page_idx / nb_pages,
                _("OCR on {doc_id} p{page_idx}").format(
                    doc_id=doc_id, page_idx=(page_idx + 1)
                )
            )
            promise = promise.then(lambda *args, **kwargs: None)
            promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
                self.core, self.core.call_all,
                args=("ocr_page_by_url", doc_url, page_idx,)
            ))
            promise = promise.then(lambda *args, **kwargs: None)

        promise = promise.then(
            self.core.call_all, "on_progress", "redo_ocr", 1.0
        )
        promise = promise.then(lambda *args, **kwargs: None)

        promise = promise.then(self.core.call_success(
            "transaction_simple_promise", [('upd', doc_id),]
        ))
        self.core.call_success("transaction_schedule", promise)



import openpaperwork_core

from ... import _


class ExtraTextRenderer(object):
    def __init__(self, core):
        self.core = core
        self.parent = None

    def get_preview_output(
                self, doc_id, doc_url, terminal_size=(80, 25),
                page_idx=0
            ):
        if self.parent is not None:
            return self.parent.get_preview_output(
                doc_id, doc_url, terminal_size, page_idx
            )
        return []

    def get_doc_output(self, doc_id, doc_url, terminal_size=(80, 25)):
        out = []
        if self.parent is not None:
            out = self.parent.get_doc_output(
                doc_id, doc_url, terminal_size
            )
        extra_text = []
        self.core.call_all("doc_get_extra_text_by_url", extra_text, doc_url)
        extra_text = "\n".join(extra_text).strip().split("\n")

        if len(extra_text) > 0:
            extra_text = ["", "  " + _("Additional text:")] + extra_text

        return extra_text + out

    def get_page_output(
                self, doc_id, doc_url, page_nb, terminal_size=(80, 25)
            ):
        if self.parent is not None:
            return self.parent.get_page_output(
                doc_id, doc_url, page_nb, terminal_size
            )
        return []

    def get_doc_infos(self, doc_id, doc_url):
        if self.parent is not None:
            return self.parent.get_doc_infos(doc_id, doc_url)
        return {}

    def get_page_infos(self, doc_id, doc_url, page_nb):
        if self.parent is not None:
            return self.parent.get_page_infos(doc_id, doc_url, page_nb)
        return {}


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 11000

    def get_interfaces(self):
        return ['doc_renderer']

    def get_deps(self):
        return [
            {
                'interface': 'extra_text',
                'defaults': ['paperwork_backend.model.extra_text'],
            },
        ]

    def doc_renderer_get(self, out):
        r = ExtraTextRenderer(self.core)
        if len(out) > 0:
            r.parent = out[-1]
        out.append(r)



import collections

import openpaperwork_core.promise


class FileImport(object):
    """
    Used as both input and output for importer objects.
    Users may import many files at once and files may not be imported at all
    (for example if they were already imported) --> FileImport objects indicate
    what must still be done, what has been done, and what has not been done at
    all.
    """
    def __init__(self, file_uris_to_import, active_doc_id=None, data=None):
        self.active_doc_id = active_doc_id
        # those attributes will be updated by importers
        self.ignored_files = list(file_uris_to_import)
        self.imported_files = set()
        self.new_doc_ids = set()
        self.upd_doc_ids = set()
        self.stats = collections.defaultdict(lambda: 0)


class BaseFileImporter(object):
    def __init__(self, core, file_import, single_file_importer_factory):
        self.core = core
        self.file_import = file_import
        self.single_file_importer_factory = single_file_importer_factory
        self.to_import = list(self._get_importables())

    def get_name(self):
        return self.single_file_importer_factory.get_name()

    def can_import(self):
        return len(self.to_import) > 0

    def get_required_data(self):
        """
        Possible requirements:
        - "password": Document is password-protected. We need the password
          to import it.

        Returns:
            {file_uri: {requirement_a, requirement_b, etc}}
        """
        out = {}
        for (orig_url, file_url) in self.to_import:
            required = self.single_file_importer_factory.get_required_data(
                file_url
            )
            if len(required) <= 0:
                continue
            out[file_url] = required
        return out

    @staticmethod
    def _upd_file_import(imported, file_import, orig_uri, file_uri):
        try:
            file_import.ignored_files.remove(orig_uri)
        except ValueError:
            pass
        if imported:
            file_import.imported_files.add(file_uri)
        else:
            file_import.ignored_files.append(file_uri)

    def _make_transactions(self, file_import):
        transactions = []
        self.core.call_all(
            "doc_transaction_start", transactions,
            len(file_import.new_doc_ids) + len(file_import.upd_doc_ids)
        )
        transactions.sort(key=lambda transaction: -transaction.priority)
        return transactions

    def _do_transactions(self, transactions, file_import):
        for doc_id in file_import.new_doc_ids:
            for transaction in transactions:
                transaction.add_doc(doc_id)
        for doc_id in file_import.upd_doc_ids:
            for transaction in transactions:
                transaction.upd_doc(doc_id)
        for transaction in transactions:
            transaction.commit()

    def get_import_promise(self, data=None):
        """
        Return a promise with all the steps required to import files
        specified in `file_import` (see constructor), transactions included.

        Must be scheduled with 'transaction_manager.transaction_schedule()'.

        Arguments:
            data are extra data that may be required
            (see `get_required_data()`)
        """
        if data is None:
            data = {}

        promise = openpaperwork_core.promise.Promise(self.core)

        for (orig_uri, file_uri) in self.to_import:
            promise = promise.then(
                self.single_file_importer_factory.make_importer(
                    self.file_import, file_uri, data
                ).get_promise()
            )
            promise = promise.then(
                self._upd_file_import, self.file_import, orig_uri, file_uri
            )

        promise = promise.then(self._make_transactions, self.file_import)
        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, self._do_transactions, args=(self.file_import,)
            )
        )

        return promise.then(
            self.core.call_all, "on_import_done", self.file_import
        )


class DirectFileImporter(BaseFileImporter):
    """
    Designed to import only explicitly selected files
    """
    def _get_importables(self):
        factory = self.single_file_importer_factory
        for file_uri in self.file_import.ignored_files:
            if factory.is_importable(self.core, file_uri):
                yield (file_uri, file_uri)


class RecursiveFileImporter(BaseFileImporter):
    """
    Assume files to import are actually directories. Look inside those
    directories to find files to import.
    """
    def _get_importables(self):
        factory = self.single_file_importer_factory
        for dir_uri in self.file_import.ignored_files:
            if not self.core.call_success("fs_isdir", dir_uri):
                continue
            for file_uri in self.core.call_success("fs_recurse", dir_uri):
                if factory.is_importable(self.core, file_uri):
                    yield (dir_uri, file_uri)

    def get_name(self):
        return self.single_file_importer_factory.get_recursive_name()



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2019  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import sys

import openpaperwork_core

from openpaperwork_core.cmd.util import ask_confirmation

from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "doc_labels",
                "defaults": ["paperwork_backend.model.labels"],
            },
            {
                "interface": "document_storage",
                "defaults": ["paperwork_backend.model.workdir"],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        label_parser = parser.add_parser(
            'label', help=_("Commands to manage labels")
        )
        subcmd_parser = label_parser.add_subparsers(
            help=_("label command"), dest='sub_command', required=True
        )

        subcmd_parser.add_parser('list')
        p = subcmd_parser.add_parser('show')
        p.add_argument(
            'doc_ids', nargs='*', default=[],
            help=_("Target documents")
        )

        p = subcmd_parser.add_parser('add')
        p.add_argument('doc_id', help=_("Target document"))
        p.add_argument('label_name', help=_("Label to add"))
        p.add_argument(
            '--color', '-c', default=None,
            help=_("Label color (ex: '#aa22cc')"), required=False
        )

        p = subcmd_parser.add_parser('remove')
        p.add_argument('doc_id', help=_("Target document"))
        p.add_argument('label_name', help=_("Label to remove"))

        p = subcmd_parser.add_parser('delete')
        p.add_argument(
            'label_name', help=_("Label to delete from *all* documents")
        )

    def _load_all_labels(self, console):
        console.print(_("Loading all labels ... "))
        promises = []
        self.core.call_all("label_load_all", promises)
        promise = promises[0]
        for p in promises[1:]:
            promise = promise.then(p)
        self.core.call_one("mainloop_schedule", promise.schedule)
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")
        console.print(_("Done"))

    def _upd_doc(self, console, doc_id):
        self.core.call_success("transaction_simple", [("upd", doc_id),])
        self.core.call_success("mainloop_quit_graceful")
        self.core.call_success("mainloop")

    def _show(self, console, doc_ids):
        out = {}
        for doc_id in doc_ids:
            doc_url = self.core.call_success("doc_id_to_url", doc_id)
            labels = set()
            self.core.call_all("doc_get_labels_by_url", labels, doc_url)
            labels = list(labels)
            labels.sort()
            out[doc_id] = labels

            labels_str = self.core.call_success(
                "format_labels", labels, separator=" "
            )
            console.print(f"{doc_id}: {labels_str}")
        return out

    def cmd_run(self, console, args):
        if args.command != 'label':
            return None

        if args.sub_command == 'list':
            self._load_all_labels(console)

            labels = set()
            self.core.call_all("labels_get_all", labels)
            labels = list(labels)
            labels.sort()

            console.print("")
            labels_str = self.core.call_success(
                "format_labels", labels, separator=" "
            )
            console.print(labels_str)
            return labels

        elif args.sub_command == 'show':

            return self._show(console, args.doc_ids)

        elif args.sub_command == "add":

            color = args.color
            if color is not None:
                # make sure the color is valid
                color = self.core.call_success("label_color_to_rgb", color)
                color = self.core.call_success("label_color_from_rgb", color)

            self._load_all_labels(console)
            doc_url = self.core.call_success("doc_id_to_url", args.doc_id)
            self.core.call_success(
                "doc_add_label_by_url", doc_url, args.label_name, color
            )

            self._upd_doc(console, args.doc_id)

            return self._show(console, [args.doc_id])

        elif args.sub_command == "remove":

            doc_url = self.core.call_success("doc_id_to_url", args.doc_id)
            self.core.call_all(
                "doc_remove_label_by_url", doc_url, args.label_name
            )
            self._upd_doc(console, args.doc_id)
            return self._show(console, [args.doc_id])

        elif args.sub_command == "delete":

            label = args.label_name

            r = ask_confirmation(
                console,
                _(
                    "Are you sure you want to delete label '%s' from all"
                    " documents ?"
                ) % label,
                default_interactive='n',
                default_non_interactive='y',
            )
            if r != 'y':
                sys.exit(1)

            all_docs = []
            self.core.call_all("storage_get_all_docs", all_docs)

            updated_docs = []

            for (doc_id, doc_url) in all_docs:
                labels = set()
                self.core.call_all("doc_get_labels_by_url", labels, doc_url)
                labels = {l for (l, c) in labels}
                if label not in labels:
                    continue

                if self.interactive:
                    print("Removing label '{}' from document '{}'".format(
                        label, doc_id
                    ))
                updated_docs.append(doc_id)
                self.core.call_all("doc_remove_label_by_url", doc_url, label)

            if self.interactive:
                sys.stdout.write("Committing changes in index ... ")
                sys.stdout.flush()

            self.core.call_success("transaction_simple", [
                ("upd", doc_id) for doc_id in updated_docs
            ])
            self.core.call_success("mainloop_quit_graceful")
            self.core.call_success("mainloop")

            if self.interactive:
                sys.stdout.write("Done\n")
            return updated_docs

        if self.interactive:
            print("Unknown label command: {}".format(args.sub_command))
        return None



"""
Workaround for the fact that "import pkg_resources" never works when frozen
with Msys2 + cx_freeze ...
"""

import logging
import os
import pathlib
import sys

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class FakePathCtxManager:
    def __init__(self, file_path):
        self.file_path = file_path

    def __enter__(self):
        return self.file_path

    def __exit__(self, type, value, traceback):
        pass


class Plugin(PluginBase):
    PRIORITY = 100

    def get_interfaces(self):
        return ['resources']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
        ]

    def resources_get_file(self, pkg, filename):
        if not getattr(sys, 'frozen', False):
            return None

        # cx_freeze: pkg_resources doesn't work (can't import it), so we keep
        # the data files beside the executable.
        path = os.path.join(
            os.path.dirname(sys.executable),
            "data",
            pkg.replace(".", os.path.sep),
            filename
        )
        if not os.path.exists(path):
            LOGGER.warning(
                "Failed to find %s/%s (tried %s)",
                pkg, filename, path
            )
            return None
        return FakePathCtxManager(pathlib.Path(path).resolve())

    def resources_get_dir(self, pkg, dirname):
        return self.resources_get_file(pkg, dirname)



import logging
import re

import openpaperwork_core

from . import util


LOGGER = logging.getLogger(__name__)

PAGE_FILENAME_PNG_FMT = "paper.{}.png"
PAGE_FILENAME_JPG_FMT = "paper.{}.jpg"
PAGE_FILENAME_FMTS = (
    PAGE_FILENAME_PNG_FMT,
    PAGE_FILENAME_JPG_FMT,
)
PAGE_FILENAME_REGEX = re.compile(
    r"paper\.(\d+)\.(jpg|png)", flags=re.IGNORECASE
)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def get_interfaces(self):
        return [
            "doc_type",
            "page_img",
            'pages',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                # to provide doc_get_nb_pages_by_url()
                'interface': 'nb_pages',
                'defaults': ['paperwork_backend.model'],
            },
        ]

    def init(self, core):
        self.core = core
        setting = self.core.call_success(
            "config_build_simple", "model",
            "img_format", lambda: "PNG"
        )
        self.core.call_all(
            "config_register", "model_img_format", setting
        )

    def is_doc(self, doc_url):
        for filename_fmt in PAGE_FILENAME_FMTS:
            page_url = self.core.call_success(
                "fs_join", doc_url, filename_fmt.format(1)
            )
            if self.core.call_success("fs_exists", page_url) is not None:
                return True
        return None

    def doc_internal_get_mtime_by_url(self, out: list, doc_url):
        mtime = util.get_doc_mtime(self.core, doc_url, PAGE_FILENAME_REGEX)
        if mtime is not None:
            out.append(mtime)

    def page_internal_get_mtime_by_url(self, out: list, doc_url, page_idx):
        for filename_fmt in PAGE_FILENAME_FMTS:
            mtime = util.get_page_mtime(
                self.core, doc_url, page_idx, filename_fmt
            )
            if mtime is not None:
                out.append(mtime)
                break

    def page_internal_get_hash_by_url(self, out: list, doc_url, page_idx):
        for filename_fmt in PAGE_FILENAME_FMTS:
            h = util.get_page_hash(self.core, doc_url, page_idx, filename_fmt)
            if h is not None:
                out.append(h)
                break

    def doc_internal_get_nb_pages_by_url(self, out: list, doc_url):
        nb_pages = util.get_nb_pages(self.core, doc_url, PAGE_FILENAME_REGEX)
        if nb_pages is None:
            return
        out.append(nb_pages)

    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
        if write:
            self.core.call_success("fs_mkdir_p", doc_url)
        for filename_fmt in PAGE_FILENAME_FMTS:
            page_url = self.core.call_success(
                "fs_join", doc_url, filename_fmt.format(page_idx + 1)
            )
            if self.core.call_success("fs_exists", page_url) is not None:
                return page_url
        if write:
            img_fmt = self.core.call_success("config_get", "model_img_format")
            if img_fmt == "PNG":
                filename_fmt = PAGE_FILENAME_PNG_FMT
            else:
                filename_fmt = PAGE_FILENAME_JPG_FMT
            return self.core.call_success(
                "fs_join", doc_url, filename_fmt.format(page_idx + 1)
            )
        return None

    def page_delete_by_url(self, doc_url, page_idx):
        r = None
        for filename_fmt in PAGE_FILENAME_FMTS:
            r = util.delete_page_file(
                self.core, filename_fmt, doc_url, page_idx
            ) or r
        return r

    def page_move_by_url(
                self,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ):
        r = None
        for filename_fmt in PAGE_FILENAME_FMTS:
            r = util.move_page_file(
                self.core, filename_fmt,
                source_doc_url, source_page_idx,
                dest_doc_url, dest_page_idx
            ) or r
        return r



#    Paperwork - Using OCR to grep dead trees the easy way
#    Copyright (C) 2012-2022  Jerome Flesch
#
#    Paperwork is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Paperwork is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
import logging

import rich.text

import openpaperwork_core
import openpaperwork_core.promise

import paperwork_backend.docimport

from .. import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                "interface": "fs",
                "defaults": "openpaperwork_gtk.fs.gio",
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                "interface": "import",
                "defaults": [
                    'paperwork_backend.docimport.img',
                    'paperwork_backend.docimport.pdf',
                ],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'import', help=_(
                "Import file(s)"
            )
        )
        p.add_argument(
            '--doc_id', '--doc', '-d', type=str, required=False,
            help=_("Target document for import"),
        )
        p.add_argument(
            '--password', type=str, required=False,
            help=_("PDF password"),
        )
        p.add_argument(
            'files', type=str, nargs='*',
            help=_("Files to import")
        )

    def _file_import_to_dict(self, file_import):
        return {
            "imported": list(file_import.imported_files),
            "ignored": list(file_import.ignored_files),
            "new_docs": list(file_import.new_doc_ids),
            "upd_docs": list(file_import.upd_doc_ids),
            "stats": dict(file_import.stats)
        }

    def cmd_run(self, console, args):
        if args.command != 'import':
            return None

        file_import = paperwork_backend.docimport.FileImport(
            [self.core.call_success("fs_safe", f) for f in args.files],
            active_doc_id=args.doc_id
        )

        importers = []
        self.core.call_all("get_importer", importers, file_import)
        choice = 0

        if len(importers) <= 0:
            console.print(
                _("Don't know how to import file(s) %s") % args.files
            )
            LOGGER.warning("Don't know how to import file(s) %s", args.files)
            return self._file_import_to_dict(file_import)

        if len(importers) > 1:
            console.print(
                _("Found many ways to import file(s) %s.") % args.files
            )
            console.print(
                _("Please select the way you want:")
            )
            choice = -1
            while choice not in range(0, len(importers)):
                for (idx, importer) in enumerate(importers):
                    console.print(f"  {idx + 1} - {importer.get_name()}")
                choice = console.input("? ")
                if choice is None:
                    return self._file_import_to_dict(file_import)
                choice = int(choice) - 1

        importer = importers[choice]
        del importers

        # We must load the labels before importing. Because the label
        # guesser may want to add labels on documents, and therefore
        # we need to know their color
        # TODO(Jflesch): That's slow and overkill. There should be a better
        # way (maybe storing the labels in ~/.local/share/paperwork2 ?)
        console.print(_("Loading labels ..."))
        promises = []
        self.core.call_all("label_load_all", promises)
        promise = promises[0]
        for p in promises[1:]:
            promise = promise.then(p)
        # use transaction_schedule to make sure that document imports
        # are not done before we have loaded the labels
        self.core.call_one("transaction_schedule", promise)

        data = {}
        if args.password is not None:
            data['password'] = args.password

        promise = importer.get_import_promise(data)

        console.print(_("Importing %s ...") % args.files)
        self.core.call_success(
            "mainloop_schedule", self.core.call_success,
            "transaction_schedule", promise
        )
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_success("mainloop")
        console.print(rich.text.Text(_("Done")))
        console.print(rich.text.Text(_("Import result:")))
        console.print(rich.text.Text(
            _("- Imported files: %s") % file_import.imported_files
        ))
        console.print(rich.text.Text(
            _("- Non-imported files: %s") % file_import.ignored_files
        ))
        console.print(rich.text.Text(
            _("- New documents: %s") % file_import.new_doc_ids
        ))
        console.print(rich.text.Text(
            _("- Updated documents: %s") % file_import.upd_doc_ids
        ))
        for (k, v) in file_import.stats.items():
            console.print(rich.text.Text(f"- {k}: {v}"))

        return self._file_import_to_dict(file_import)



import logging

import openpaperwork_core
import openpaperwork_core.promise

from . import (
    DirectFileImporter,
    FileImport,
    RecursiveFileImporter
)
from .. import _


LOGGER = logging.getLogger(__name__)


class SingleDocImporter(object):
    def __init__(self, plugin, file_import, src_file_uri):
        self.plugin = plugin
        self.core = plugin.core
        self.file_import = file_import
        self.src_file_uri = src_file_uri
        self.doc_id = None
        self.doc_url = None

    def _basic_import(self, file_url):
        file_hash = self.core.call_success("fs_hash", file_url)
        other_doc_id = self.core.call_success(
            "index_get_doc_id_by_hash", file_hash
        )
        if other_doc_id is not None:
            LOGGER.info("%s has already been imported", file_url)
            self.file_import.stats[_("Already imported")] += 1
            return False

        LOGGER.info("Importing %s", file_url)
        (self.doc_id, self.doc_url) = self.core.call_success(
            "doc_convert_and_import", file_url
        )
        self.file_import.new_doc_ids.add(self.doc_id)
        self.file_import.stats[_("Documents")] += 1

        mime = self.core.call_success("fs_get_mime", file_url)
        if mime is not None:
            self.file_import.stats[self.plugin.file_types_by_mime[mime]] += 1
        elif "." in file_url:
            file_ext = file_url.rsplit(".", 1)[-1].lower()
            self.file_import.stats[
                self.plugin.file_types_by_ext[file_ext][1]
            ] += 1

        return True

    def get_promise(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self._basic_import, args=(self.src_file_uri,)
        )


class SingleDocImporterFactory(object):
    def __init__(self, plugin):
        self.plugin = plugin
        self.core = plugin.core

    @staticmethod
    def get_name():
        return _("Import office document")

    @staticmethod
    def get_recursive_name():
        return _("Import office documents recursively")

    def is_importable(self, core, file_url):
        mime = core.call_success("fs_get_mime", file_url)
        if mime is not None:
            return mime in self.plugin.file_types_by_mime
        if "." not in file_url:
            return False
        file_ext = file_url.rsplit(".", 1)[-1].lower()
        return file_ext in self.plugin.file_types_by_ext

    def get_required_data(self, file_uri):
        return set()

    def make_importer(self, file_import, file_uri, data):
        return SingleDocImporter(self.plugin, file_import, file_uri)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.file_types_by_ext = {}
        self.file_types_by_mime = {}

    def get_interfaces(self):
        return ["import"]

    def get_deps(self):
        return [
            {
                "interface": "doc_convert_and_import",
                "defaults": ["paperwork_backend.model.converted"],
            },
            {
                "interface": "doc_converter",
                "defaults": ["paperwork_backend.converter.libreoffice"],
            },
            {
                "interface": "fs",
                "defaults": ["openpaperwork_gtk.fs.gio"],
            },
            {
                "interface": "mainloop",
                "defaults": ["openpaperwork_gtk.mainloop.glib"],
            },
            {
                "interface": "thread",
                "defaults": ["openpaperwork_core.thread.simple"],
            },
        ]

    def init(self, core):
        super().init(core)
        file_types = []
        self.core.call_all("converter_get_file_types", file_types)

        self.file_types_by_ext = {}
        for (mime, ext, human_name) in file_types:  # order matters !
            if ext in self.file_types_by_ext:
                continue
            self.file_types_by_ext[ext] = (mime, human_name)

        self.file_types_by_mime = {}
        for (mime, ext, human_name) in file_types:  # order matters !
            if mime in self.file_types_by_mime:
                continue
            self.file_types_by_mime[mime] = human_name

    def get_import_mime_types(self, out: list):
        for (mime, human_desc) in self.file_types_by_ext.values():
            out.add((human_desc, mime))
        if len(self.file_types_by_ext) > 0:
            out.add((_("Office document folder"), "inode/directory"))

    def get_importer(self, out: list, file_import: FileImport):
        importer = DirectFileImporter(
            self.core, file_import, SingleDocImporterFactory(self)
        )
        if importer.can_import():
            out.append(importer)

        importer = RecursiveFileImporter(
            self.core, file_import, SingleDocImporterFactory(self)
        )
        if importer.can_import():
            out.append(importer)



import gettext

import openpaperwork_core
import openpaperwork_gtk


def _(s):
    return gettext.dgettext('paperwork_backend', s)


DEFAULT_CONFIG_PLUGINS = openpaperwork_core.MINIMUM_CONFIG_PLUGINS + [
    'paperwork_backend.app',
]

DEFAULT_PLUGINS = (
    openpaperwork_core.RECOMMENDED_PLUGINS +
    openpaperwork_gtk.CLI_PLUGINS +
    [
        'openpaperwork_core.beacon.stats',
        'openpaperwork_core.beacon.sysinfo',
        'openpaperwork_core.bug_report.censor',
        'openpaperwork_core.censor',
        'openpaperwork_core.pillow.img',
        'openpaperwork_core.pillow.util',
        'paperwork_backend.authors',
        'paperwork_backend.authors.translators',
        'paperwork_backend.beacon.update',
        'paperwork_backend.cairo.blur',
        'paperwork_backend.cairo.cache',
        'paperwork_backend.cairo.pillow',
        'paperwork_backend.cairo.poppler',
        'paperwork_backend.chkworkdir.corrupted_page_map',
        'paperwork_backend.chkworkdir.empty_doc',
        'paperwork_backend.chkworkdir.file_at_workdir_root',
        'paperwork_backend.chkworkdir.label_color',
        'paperwork_backend.converter.libreoffice',
        'paperwork_backend.datadirhandler',
        'paperwork_backend.docexport.automatic',
        'paperwork_backend.docexport.generic',
        'paperwork_backend.docexport.img',
        'paperwork_backend.docexport.pdf',
        'paperwork_backend.docexport.pillowfight',
        'paperwork_backend.docimport.converted',
        'paperwork_backend.docimport.img',
        'paperwork_backend.docimport.pdf',
        'paperwork_backend.docscan.libinsane',
        'paperwork_backend.docscan.scan2doc',
        'paperwork_backend.doctracker',
        # ACE is disabled by default: it's slow, and actually makes some scans
        # harder to read.
        # 'paperwork_backend.guesswork.color.libpillowfight',
        'paperwork_backend.guesswork.label.sklearn',
        'paperwork_backend.guesswork.ocr.pyocr',
        'paperwork_backend.guesswork.orientation.pyocr',
        'paperwork_backend.i18n.pycountry',
        'paperwork_backend.i18n.scanner',
        'paperwork_backend.imgedit.color',
        'paperwork_backend.imgedit.crop',
        'paperwork_backend.imgedit.rotate',
        'paperwork_backend.index.whoosh',
        'paperwork_backend.l10n',
        'paperwork_backend.model',
        'paperwork_backend.model.converted',
        'paperwork_backend.model.extra_text',
        'paperwork_backend.model.hocr',
        'paperwork_backend.model.img',
        'paperwork_backend.model.img_overlay',
        'paperwork_backend.model.labels',
        'paperwork_backend.model.pdf',
        'paperwork_backend.model.thumbnail',
        'paperwork_backend.model.workdir',
        'paperwork_backend.pageedit.pageeditor',
        'paperwork_backend.pagetracker',
        'paperwork_backend.pillow.pdf',
        'paperwork_backend.poppler.file',
        'paperwork_backend.poppler.memory',
        'paperwork_backend.pyocr',
        'paperwork_backend.sync',
    ]
)



import logging

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise


try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

    # workaround so chkdeps can still be called
    class GObject(object):
        TYPE_BOOLEAN = 0

        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass

from .... import _


LOGGER = logging.getLogger(__name__)


class Page(GObject.GObject):
    __gsignals__ = {
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'visibility_changed': (GObject.SignalFlags.RUN_LAST, None, (
            GObject.TYPE_BOOLEAN,
        )),
    }

    def __init__(self, core, doc_id, doc_url, page_idx, nb_pages):
        super().__init__()
        self.core = core
        self.doc_id = doc_id
        self.doc_url = doc_url
        self.page_idx = page_idx
        self.nb_pages = nb_pages
        self.visible = False

        self.mtime = 0

        self.zoom = 1.0

        self.widget_tree = None
        self.widget = None
        self.renderer = None

        self._load_renderer()
        self.rebuild_widget()

    def _load_renderer(self):
        page_img_url = self.core.call_success(
            "page_get_img_url", self.doc_url, self.page_idx
        )
        LOGGER.info(
            "URL for %s p%d: %s", self.doc_url, self.page_idx, page_img_url
        )
        assert page_img_url is not None

        self.renderer = self.core.call_success(
            "cairo_renderer_by_url", "page_loader", page_img_url
        )
        self.renderer.connect("getting_size", self._on_renderer_getting_size)
        self.renderer.connect("size_obtained", self._on_renderer_size)
        self.renderer.connect("img_obtained", self._on_renderer_img)

    def rebuild_widget(self):
        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageview", "pageview.glade"
        )
        self.widget = self.widget_tree.get_object("pageview_area")
        self.widget.set_visible(False)  # visible in the GTK sense
        self.widget.connect("draw", self._on_draw)
        self.resize()

    def __str__(self):
        return "{} p{}".format(self.doc_id, self.page_idx)

    def _on_renderer_getting_size(self, renderer):
        self.core.call_all(
            "on_progress", "loading_page_sizes",
            self.page_idx / self.nb_pages,
            _("Loading page {}/{} ...").format(self.page_idx, self.nb_pages)
        )
        self.emit('getting_size')

    def _on_renderer_size(self, renderer):
        LOGGER.info(
            "Page %d: size %d x %d",
            self.page_idx, self.renderer.size[0], self.renderer.size[1]
        )
        self.widget.set_visible(True)
        self.resize()
        self.core.call_all("on_page_size_obtained", self)
        self.emit('size_obtained')

    def _on_renderer_img(self, renderer):
        assert self.visible
        self.refresh()
        self.core.call_all("on_page_img_obtained", self)
        self.emit('img_obtained')

    def load(self):
        if not self.refresh(reload=True):
            # renderer has already loaded the page --> reemit the page size
            if self.renderer.size[0] != 0:
                self.core.call_success(
                    "mainloop_schedule", self._on_renderer_size, self.renderer
                )

    def close(self):
        self.renderer.close()

    def get_zoom(self):
        return self.zoom

    def set_zoom(self, zoom):
        self.zoom = zoom
        self.resize()

    def get_full_size(self):
        return self.renderer.size

    def get_size(self):
        return (
            int(self.renderer.size[0] * self.zoom),
            int(self.renderer.size[1] * self.zoom)
        )

    def resize(self):
        if self.renderer.size[0] == 0:
            return
        self.renderer.zoom = self.zoom
        size = self.get_size()
        self.widget.set_size_request(size[0], size[1])

    def refresh(self, reload=False):
        if reload:
            # only if the mtime has changed. Otherwise there is no point.
            mtime = self.core.call_success(
                "page_get_mtime_by_url", self.doc_url, self.page_idx
            )
            if self.mtime == mtime:
                reload = False
            else:
                self.mtime = mtime

        if reload:
            self.set_visible(False)
            self.close()
            self._load_renderer()
            self.renderer.start()
            return True
        elif self.widget is not None:
            self.widget.queue_draw()
            return False

    def hide(self):
        self.visible = False
        self.renderer.hide()
        self.core.call_all("on_page_visibility_changed", self, False)
        self.emit('visibility_changed', False)

    def show(self):
        self.visible = True
        self.renderer.render()
        self.core.call_all("on_page_visibility_changed", self, True)
        self.emit('visibility_changed', True)

    def set_visible(self, visible):
        if self.visible == visible:
            return

        if visible:
            self.show()
        else:
            self.hide()

    def get_visible(self):
        return self.visible

    def _on_draw(self, widget, cairo_ctx):
        self.renderer.draw(cairo_ctx)
        self.core.call_all("on_page_draw", cairo_ctx, self)

    def blur(self):
        self.renderer.blur()
        self.widget.queue_draw()

    def unblur(self):
        self.renderer.unblur()
        self.widget.queue_draw()

    def detach_from_parent(self):
        parent = self.widget.get_parent()
        if parent is None:
            return
        parent.remove(self.widget)


if GLIB_AVAILABLE:
    GObject.type_register(Page)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def __init__(self):
        super().__init__()
        self.pages = []
        self.nb_to_load = 0
        self.active_doc = (None, None)

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_pageview',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'cairo_url',
                'defaults': [
                    'paperwork_backend.cairo.pillow',
                    'paperwork_backend.cairo.poppler',
                ],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'page_img',
                'defaults': [
                    'paperwork_backend.model.img',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def init(self, core):
        super().init(core)
        self.core.call_all(
            "work_queue_create", "page_loader", stop_on_quit=True
        )

    def doc_close(self):
        self.core.call_success("work_queue_cancel_all", "page_loader")
        for page in self.pages:
            page.close()
        self.pages = []

    def doc_open_components(self, out: list, doc_id, doc_url):
        active_doc = (doc_id, doc_url)
        if self.active_doc != active_doc:
            self.doc_close()
            self.active_doc = active_doc

        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        LOGGER.info("Number of pages displayed: %s", nb_pages)
        if nb_pages is None:
            LOGGER.warning("Failed to get the number of pages in %s", doc_id)
            nb_pages = 0

        self.core.call_all("on_objref_graph")

        self.core.call_all(
            "on_perfcheck_start",
            "pageview->doc_open_components({})".format(doc_id)
        )

        # drop the extra pages we have if any
        for page in self.pages[nb_pages:]:
            page.close()
        self.pages = self.pages[:nb_pages]

        # reuse the pages we already have to avoid useless refreshes
        for page in self.pages:
            page.detach_from_parent()

        # add any new page
        for page_idx in range(len(self.pages), nb_pages):
            self.pages.append(Page(
                self.core, doc_id, doc_url, page_idx, nb_pages
            ))

        LOGGER.info(
            "%d pages in the documents (%d components)",
            nb_pages, len(self.pages)
        )

        self.nb_to_load = len(self.pages)
        for page in self.pages:
            page.connect("size_obtained", self._on_page_img_size_obtained)
            out.append((True, page))

        self.core.call_all(
            "on_perfcheck_stop",
            "pageview->doc_open_components({})".format(doc_id),
            nb_pages=nb_pages
        )

    def _on_page_img_size_obtained(self, page):
        self.nb_to_load -= 1
        if self.nb_to_load > 0:
            return
        self.nb_to_load = 0
        self.core.call_all("on_progress", "loading_page_sizes", 1.0)

    def pageview_refresh_all(self):
        for page in self.pages:
            page.refresh()

    def on_screenshot_before(self):
        LOGGER.info("Blurring pages")
        for page in self.pages:
            page.blur()

    def on_screenshot_after(self):
        LOGGER.info("Unblurring pages")
        for page in self.pages:
            page.unblur()



import random

import fabulous.image
import fabulous.text

import rich.text

import openpaperwork_core

from ... import _


# XXX(Jflesch): crappy workaround for an unmaintained library ...
fabulous.image.basestring = str


VALIDATED_FONTS = """
comic
Comic_Sans_MS
Comic_Sans_MS_Bold
comicbd
LiberationMono-Bold
LiberationSans-Bold
times
timesbd
FreeMono
FreeMonoBold
FreeMonoBoldOblique
FreeMonoOblique
FreeSans
FreeSansBold
FreeSansBoldOblique
FreeSansOblique
FreeSerif
FreeSerifBold
FreeSerifBoldItalic
FreeSerifItalic
""".strip().split()  # to make copy and paste faster

COLORS = [
    '#0099ff',
    '#ff4400',
    '#00bd22',
    '#bf00ff',
]


class Paperwork(object):
    def __init__(self, core, fonts):
        self.core = core
        self.fonts = fonts

    def show(self, console):
        nb_lines = 0

        logo_ctx = self.core.call_success(
            "resources_get_file",
            "paperwork_shell.cmd.about", "logo.png"
        )
        with logo_ctx as logo:
            logo = fabulous.image.Image(logo, width=60)
        logo = logo.reduce(logo.convert())
        for line in logo:
            console.print(rich.text.Text.from_ansi((16 * " ") + line))
            nb_lines += 1

        font = "FreeSans"
        if font not in self.fonts:
            font = self.fonts[0]

        txt = fabulous.text.Text(
            "Paperwork",
            skew=5, color="#abcdef", font=font, fsize=25, shadow=False
        )
        for line in txt:
            console.print(rich.text.Text.from_ansi(line))
            nb_lines += 1

        console.print("Paperwork")
        console.print(
            (8 * " ")
            + _('Version: ') + self.core.call_success("app_get_version")
        )
        console.print(
            (8 * " ")
            + _("Because sorting documents is a machine's job.")
        )
        nb_lines += 3
        return nb_lines


class Section(object):
    def __init__(self, name, authors, fonts):
        self.name = name
        self.authors = authors
        self.fonts = fonts

    @staticmethod
    def _group_small_words(words):
        buf = []
        for word in words:
            buf.append(word)
            if len(word) > 3:
                yield " ".join(buf)
                buf = []
        if len(buf) > 0:
            yield " ".join(buf)

    def show(self, console):
        nb_lines = 0
        color = random.choice(COLORS)
        font = random.choice(self.fonts)

        words = self.name.split()
        for word in self._group_small_words(words):
            txt = fabulous.text.Text(
                word, skew=0, color=color, font=font, fsize=18, shadow=False
            )
            for line in txt:
                console.print(rich.text.Text.from_ansi(line))
                nb_lines += 1

        console.print(self.name)
        nb_lines += 1
        for author in self.authors:
            txt = author[1]
            if author[2] > 0:
                txt += " ({})".format(author[2])
            console.print((8 * " ") + txt)
            nb_lines += 1

        return nb_lines


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                'interface': 'app',
                'defaults': ['paperwork_backend.app'],
            },
            {
                'interface': 'authors',
                'defaults': ['paperwork_backend.authors'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        parser.add_parser('about', help=_("About Paperwork"))

    def cmd_run(self, console, args):
        if args.command != 'about':
            return None

        fonts = self._get_available_fonts()

        sections = {}
        self.core.call_all("authors_get", sections)
        sections = [x for x in sections.items()]
        sections.sort(key=lambda x: x[0].lower())

        sections = (
            [Paperwork(self.core, fonts)] +
            [Section(k, v, fonts) for (k, v) in sections if len(v) >= 0]
        )

        for section in sections:
            section.show(console)
            if section != sections[-1]:
                for x in range(0, 7):
                    console.print("")
        console.print("")
        console.print("")

    def _get_available_fonts(self):
        all_fonts = fabulous.text.get_font_files()
        all_fonts = {n for n in all_fonts.keys()}

        out = []
        for font in VALIDATED_FONTS:
            if font not in all_fonts:
                continue
            out.append(font)
        return out



import logging
import multiprocessing
import queue
import threading

from . import Task
from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Thread(threading.Thread):
    def __init__(self, plugin, thread_id):
        super().__init__(name="paperwork_thread_{}".format(thread_id))
        self.daemon = True
        self.task_queue = plugin.queue
        self.core = plugin.core
        self.running = True

    def run(self):
        LOGGER.info("Thread %s ready", self.name)
        while True:
            task = self.task_queue.get()
            if task is None:
                break
            task.do()
        LOGGER.info("Thread %s stopped", self.name)


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.queue = queue.Queue()
        self.pool = None

    def get_interfaces(self):
        return ['thread']

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_core.mainloop.asyncio'],
            },
        ]

    def on_mainloop_start(self):
        if self.pool is None:
            self.pool = [
                Thread(self, x)
                for x in range(0, max(4, multiprocessing.cpu_count()))
            ]
            for t in self.pool:
                t.start()

    def on_mainloop_quit(self):
        if self.pool is not None:
            for t in self.pool:
                self.queue.put(None)
            # in case the mainloop is restarted later:
            self.queue = queue.Queue()
            self.pool = None

    def thread_start(self, func, *args, **kwargs):
        if self.pool is None:
            self.on_mainloop_start()
        task = Task(self.core, func, args, kwargs)
        self.queue.put(task)



import logging
import os
try:
    import importlib.resources
    IMPORTLIB_RESOURCES_AVAILABLE = True
except Exception:
    IMPORTLIB_RESOURCES_AVAILABLE = False

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['resources']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
        ]

    def resources_get_file(self, pkg, filename):
        if not IMPORTLIB_RESOURCES_AVAILABLE:
            LOGGER.warning("importlib.resources not available !")
            return

        path = importlib.resources.files(pkg).joinpath(filename)

        if not os.access(path, os.R_OK):
            LOGGER.warning(
                "Failed to find resource file %s/%s,"
                " tried at path %s.", pkg, filename, path
            )
            return None

        LOGGER.debug("%s:%s --> %s", pkg, filename, path)
        return importlib.resources.as_file(path)

    def resources_get_dir(self, pkg, dirname):
        return self.resources_get_file(pkg, dirname)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "page_print"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = None
        self.active_page_idx = -1
        self.active_windows = []

    def get_interfaces(self):
        return [
            'action',
            'action_page',
            'action_page_print',
            'chkdeps',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_print',
                'defaults': ['paperwork_gtk.print'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return

        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._print)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def doc_close(self):
        self.active_doc = None

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

    def _print(self, *args, **kwargs):
        assert self.active_doc is not None
        self.core.call_all(
            "doc_print", *self.active_doc, [self.active_page_idx]
        )



from typing import Any
from typing import Dict
from typing import Tuple
import logging

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

# TODO(Jflesch): bad
import paperwork_backend.model.pdf


CAIRO_AVAILABLE = False
GLIB_AVAILABLE = False

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    pass

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    # dummy so chkdeps can still be called
    class GObject(object):  # type: ignore[no-redef]
        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass


LOGGER = logging.getLogger(__name__)
DELAY = 0.01
POPPLER_DOCS: Dict[str, Tuple[Any, int]] = {}


class ImgSurface(object):
    # wrapper so it can be weakref
    def __init__(self, surface):
        self.surface = surface


class CairoRenderer(GObject.GObject):
    __gsignals__ = {
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),
    }
    OUTLINE = (0.5, 0.5, 0.5)

    def __init__(
            self, core, work_queue_name, file_url, page_idx, password=None):
        global POPPLER_DOCS

        super().__init__()
        self.core = core
        self.work_queue_name = work_queue_name
        self.file_url = file_url
        self.page_idx = page_idx
        self.password = password

        self.visible = False
        self.size = (0, 0)
        self.zoom = 1.0

        if file_url in POPPLER_DOCS:
            (doc, refcount) = POPPLER_DOCS[file_url]
        else:
            LOGGER.info("Opening PDF file {}".format(file_url))
            doc = self.core.call_success(
                "poppler_open", file_url, password=password
            )
            refcount = 0
        POPPLER_DOCS[file_url] = (doc, refcount + 1)

        self.page = doc.get_page(page_idx)

        promise = openpaperwork_core.promise.Promise(
            self.core, self.emit, args=("getting_size",)
        )
        promise = promise.then(openpaperwork_core.promise.Promise(
            self.core, self.page.get_size
        ))
        promise = promise.then(lambda size: (
            size[0] * paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
            size[1] * paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
        ))
        promise.then(self._set_size)

        if page_idx % 25 == 0:
            # Gives back a bit of CPU time to GTK so the GUI remains
            # usable, but not too much so we don't recompute the layout too
            # often
            promise = promise.then(openpaperwork_core.promise.DelayPromise(
                core, DELAY
            ))

        self.get_size_promise = promise

    def __str__(self):
        return "CairoRenderer({} p{})".format(self.file_url, self.page_idx)

    def _set_size(self, size):
        if self.page is None:
            # Document has been closed while we looked for its size
            return
        self.size = size
        self.emit('size_obtained')
        if self.visible:
            self.render(force=True)

    def start(self):
        self.core.call_success(
            "work_queue_add_promise",
            self.work_queue_name, self.get_size_promise
        )

    def render(self, force=False):
        if self.visible and not force:
            return
        self.visible = True
        if self.size == (0, 0):
            return
        self.emit('img_obtained')

    def hide(self):
        self.visible = False

    def close(self):
        global POPPLER_DOCS
        self.hide()
        self.page = None
        self.size = (0, 0)
        (doc, refcount) = POPPLER_DOCS.get(self.file_url, (None, None))
        if refcount is None:
            LOGGER.warning("Double close for %s", self.file_url)
            return
        refcount -= 1
        if refcount > 0:
            POPPLER_DOCS[self.file_url] = (doc, refcount)
            return
        LOGGER.info("Closing PDF file %s", self.file_url)
        POPPLER_DOCS.pop(self.file_url, None)

    def _draw(self, cairo_ctx, zoom):
        try:
            cairo_ctx.save()
            try:
                cairo_ctx.set_source_rgb(1.0, 1.0, 1.0)
                cairo_ctx.scale(zoom, zoom)
                cairo_ctx.rectangle(0, 0, self.size[0], self.size[1])
                cairo_ctx.scale(
                    paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
                    paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
                )
                cairo_ctx.clip()
                cairo_ctx.paint()
                self.page.render(cairo_ctx)

                cairo_ctx.scale(
                    1 / paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
                    1 / paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
                )

                cairo_ctx.set_source_rgb(*self.OUTLINE)
                outline_width = 1 / zoom
                cairo_ctx.set_line_width(outline_width)
                cairo_ctx.rectangle(
                    0, 0,
                    self.size[0] - outline_width, self.size[1] - outline_width
                )
                cairo_ctx.stroke()
            finally:
                cairo_ctx.restore()
        except Exception as exc:
            LOGGER.error("CairoRenderer.draw() failed (PDF)", exc_info=exc)
            # WORKAROUND(Jflesch): with some malformed PDF file, we get an
            # exception on ctx.restore(), but the drawing was actually done.

    def draw(self, cairo_ctx):
        if not self.visible or self.page is None or self.size[0] == 0:
            return

        task = "pdf_to_cairo_draw({}, p{})".format(
            self.file_url, self.page_idx
        )
        self.core.call_all("on_perfcheck_start", task)
        self._draw(cairo_ctx, self.zoom)
        self.core.call_all("on_perfcheck_stop", task, size=self.size)

    def blur(self):
        pass

    def unblur(self):
        pass


if GLIB_AVAILABLE:
    GObject.type_register(CairoRenderer)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000
    FILE_EXTENSION = ".pdf"

    def get_interfaces(self):
        return [
            'cairo_url',
            'chkdeps',
            'page_img_size',
            'pdf_cairo_url',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'poppler',
                'defaults': ['paperwork_backend.poppler.memory'],
            },
            {
                'interface': 'urls',
                'defaults': ['openpaperwork_core.urls'],
            },
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def _check_is_pdf(self, file_url):
        (url, args) = self.core.call_success("url_args_split", file_url)
        if not url.lower().endswith(self.FILE_EXTENSION):
            return (None, None, None)

        password = args.get('password', None)
        if password is not None:
            password = bytes.fromhex(password).decode("utf-8")
        page_idx = int(args.get('page', 1)) - 1
        return (url, page_idx, password)

    def url_to_img_size(self, file_url):
        (file_url, page_idx, password) = self._check_is_pdf(file_url)
        if file_url is None:
            return None

        task = "url_to_img_size({})".format(file_url)
        self.core.call_all("on_perfcheck_start", task)
        doc = self.core.call_success(
            "poppler_open", file_url, password=password
        )
        page = doc.get_page(page_idx)

        base_size = page.get_size()
        size = (  # scale up because default size if too small for reading
            int(base_size[0]) * paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
            int(base_size[1]) * paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
        )

        self.core.call_all("on_perfcheck_stop", task, size=size)
        return size

    def url_to_img_size_promise(self, file_url):
        (page_url, page_idx, password) = self._check_is_pdf(file_url)
        if page_url is None:
            return None

        return openpaperwork_core.promise.Promise(
            self.core, self.url_to_img_size, args=(file_url,)
        )

    def pdf_page_to_cairo_surface(self, file_url, page_idx, password=None):
        task = "pdf_page_to_cairo_surface({} p{})".format(file_url, page_idx)

        self.core.call_all("on_perfcheck_start", task)

        doc = self.core.call_success(
            "poppler_open", file_url, password=password
        )
        page = doc.get_page(page_idx)

        # WORKAROUND(Jflesch):
        # Some corrupted PDFs may cause `page = None` here. (PDF declaring
        # more pages than they actually have).
        if page is None:
            return None

        base_size = page.get_size()
        size = (  # scale up because default size if too small for reading
            int(base_size[0]) * paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
            int(base_size[1]) * paperwork_backend.model.pdf.PDF_RENDER_FACTOR,
        )

        width = int(size[0])
        height = int(size[1])
        factor_w = width / base_size[0]
        factor_h = height / base_size[1]

        surface = ImgSurface(cairo.ImageSurface(
            cairo.FORMAT_ARGB32, width, height
        ))
        self.core.call_all("on_objref_track", surface)

        try:
            ctx = cairo.Context(surface.surface)
            ctx.save()
            try:
                ctx.set_source_rgb(1.0, 1.0, 1.0)
                ctx.rectangle(0, 0, width, height)
                ctx.clip()
                ctx.paint()

                ctx.scale(factor_w, factor_h)
                page.render(ctx)
            finally:
                ctx.restore()
        except Exception as exc:
            LOGGER.error("pdf_page_to_cairo_surface() failed", exc_info=exc)
            # WORKAROUND(Jflesch): with some malformed PDF file, we get an
            # exception on ctx.restore(), but the drawing was actually done.

        self.core.call_all("on_perfcheck_stop", task, size=(width, height))
        return surface

    def url_to_cairo_surface(self, file_url):
        (file_url, page_idx, password) = self._check_is_pdf(file_url)
        if file_url is None:
            return None
        return self.pdf_page_to_cairo_surface(file_url, page_idx, password)

    def url_to_cairo_surface_promise(self, file_url):
        (file_url, page_idx, password) = self._check_is_pdf(file_url)
        if file_url is None:
            return None
        return openpaperwork_core.promise.Promise(
            self.core, self.pdf_page_to_cairo_surface,
            args=(file_url, page_idx, password)
        )

    def cairo_renderer_by_url(self, work_queue_name, file_url, **kwargs):
        (file_url, page_idx, password) = self._check_is_pdf(file_url)
        if file_url is None:
            return None
        return CairoRenderer(
            self.core, work_queue_name, file_url, page_idx, password
        )



import faulthandler
import logging
import os

from .. import (_, PluginBase)


LOGGER = logging.getLogger(__name__)
LOG_DATE_FORMAT = "%Y%m%d_%H%M_%S"
MAX_DAYS = 31


class LogLine(object):
    def __init__(self, line):
        self.line = line
        self.next = None


class LogHandler(logging.Handler):
    LOG_FORMAT = '[%(levelname)-6s] [%(name)-30s] %(message)s'
    MAX_LINES = 5000
    MAX_UNCAUGHT_LOGGUED = 20

    def __init__(self, core, archiver):
        super().__init__()
        self.core = core
        self.archiver = archiver

        self.first_line = None
        self.last_line = None
        self.nb_lines = 0

        self.nb_uncaught_loggued = 0

        self.out_file_url = self.archiver.get_new()
        self.formatter = logging.Formatter(self.LOG_FORMAT)
        self.out_fd = self.core.call_success(
            "fs_open", self.out_file_url, 'w',
            needs_fileno=True
        )
        faulthandler.disable()
        faulthandler.enable(file=self.out_fd)

    def emit(self, record):
        line = self.formatter.format(record) + "\n"
        self.out_fd.write(line)

        line = LogLine(line)
        if self.last_line is not None:
            self.last_line.next = line
        self.last_line = line

        if self.first_line is None:
            self.first_line = line

        if self.nb_lines >= self.MAX_LINES:
            self.last_line = self.last_line.next
        else:
            self.nb_lines += 1

    def log_uncaught_exception(self):
        if self.nb_uncaught_loggued >= self.MAX_UNCAUGHT_LOGGUED:
            # avoid logging too much
            return

        self.nb_uncaught_loggued += 1

        out_file_url = self.archiver.get_new(
            name="uncaught_exception_logs"
        )
        self.formatter = logging.Formatter(self.LOG_FORMAT)

        with self.core.call_success("fs_open", out_file_url, 'w') as fd:
            line = self.first_line
            while line is not None:
                fd.write(line.line)
                line = line.next


class Plugin(PluginBase):
    PRIORITY = -10000

    def __init__(self):
        super().__init__()
        logging.getLogger().setLevel(logging.INFO)
        self.archiver = None

    def get_interfaces(self):
        return [
            'bug_report_attachments',
            'log_archiver',
            'uncaught_exception_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'file_archives',
                'defaults': ['openpaperwork_core.archives'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'uncaught_exception',
                'defaults': ['openpaperwork_core.uncaught_exception'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.archiver = self.core.call_success(
            "file_archive_get", storage_name="logs", file_extension="txt"
        )
        self.log_handler = LogHandler(core, self.archiver)
        logging.getLogger().addHandler(self.log_handler)

    def on_uncaught_exception(self, exc_info):
        self.log_handler.log_uncaught_exception()

    def bug_report_get_attachments(self, inputs: dict):
        archived = list(self.archiver.get_archived())
        archived.sort(key=lambda x: x[0], reverse=True)

        for (nb, (date, file_url)) in enumerate(archived):
            inputs[file_url] = {
                'date': date,
                # Users tend to send only the pre-selected elements in the bug
                # reports. However, when they report a sudden crash, we need
                # the logs of the previous session, not just the current one.
                # --> Include the logs of the last 2 previous sessions too.
                'include_by_default': nb <= 2,
                'file_type': _("Log file"),
                'file_url': file_url,
                'file_size': self.core.call_success("fs_getsize", file_url),
            }

        LOGGER.info("Flushing logs to disk")
        self.log_handler.out_fd.flush()
        os.fsync(self.log_handler.out_fd.fileno())
        file_url = self.log_handler.out_file_url
        inputs[file_url] = {
            'date': None,  # now
            'include_by_default': True,
            'file_type': _("Log file"),
            'file_url': file_url,
            'file_size': self.core.call_success("fs_getsize", file_url),
        }



import itertools

import PIL.Image

import openpaperwork_core
import openpaperwork_core.promise


SCAN_ID_GENERATOR = itertools.count()


class Source(object):
    def __init__(self, core, scanner, source_id):
        self.core = core
        self.scanner = scanner
        self.source_id = source_id

    def __str__(self):
        return "{}:{}".format(str(self.scanner), self.source_id)

    def set_as_default(self):
        raise NotImplementedError()

    def get_resolutions_promise(self):
        def get_resolutions():
            return [25, 100, 200, 300, 400, 500, 600]
        return openpaperwork_core.promise.Promise(self.core, get_resolutions)

    def set_default_resolution(self, resolution):
        raise NotImplementedError()

    def scan(self, scan_id=None, resolution=None, max_pages=9999):
        if scan_id is None:
            scan_id = next(SCAN_ID_GENERATOR)

        test_chunk = PIL.Image.new("RGB", (100, 200), (171, 205, 239))
        test_page = PIL.Image.new("RGB", (200, 200), (171, 205, 239))

        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_feed_start", scan_id
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_page_start",
            scan_id, 0,
            None,  # TODO(Jflesch): scan params
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_chunk",
            scan_id,
            None,  # TODO(Jflesch): scan_params
            test_chunk
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_chunk",
            scan_id,
            None,  # TODO(Jflesch): scan_params
            test_chunk
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_chunk",
            scan_id,
            None,  # TODO(Jflesch): scan_params
            test_chunk
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_page_end", scan_id, 0,
            test_page
        )
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_scan_feed_end", scan_id
        )
        return (self, scan_id, [test_page])

    def scan_promise(self, *args, scan_id=None, **kwargs):
        if scan_id is None:
            scan_id = next(SCAN_ID_GENERATOR)
        kwargs['scan_id'] = scan_id
        return (scan_id, openpaperwork_core.promise.ThreadedPromise(
            self.core, self.scan, args=args, kwargs=kwargs
        ))

    def close(self, *args, **kwargs):
        pass


class Scanner(object):
    def __init__(self, core, scanner_id):
        self.core = core
        self.dev_id = scanner_id

    def __str__(self):
        return self.dev_id

    def close(self, *args, **kwargs):
        pass

    def get_sources(self):
        return {
            "fake_source0": Source(self.core, self, "fake_source0"),
            "fake_source1": Source(self.core, self, "fake_source1"),
        }

    def get_sources_promise(self):
        return openpaperwork_core.promise.Promise(self.core, self.get_sources)

    def get_source(self, source_id):
        return self.get_sources()[source_id]

    def set_as_default(self):
        raise NotImplementedError()


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ["scan"]

    def get_deps(self):
        return [
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def scan_list_scanners_promise(self):
        def list_scanners():
            return [
                ("fake:scanner0", "Super scanner #0"),
                ("fake:scanner1", "Super scanner #1"),
            ]

        return openpaperwork_core.promise.Promise(self.core, list_scanners)

    def scan_get_scanner_promise(self, dev_id):
        def get_scanner():
            if dev_id != "fake:scanner0" and dev_id != "fake:scanner1":
                return None
            return Scanner(self.core, dev_id)

        return openpaperwork_core.promise.Promise(self.core, get_scanner)

    def scan_promise(self, *args, **kwargs):
        scan_id = next(SCAN_ID_GENERATOR)

        promise = self.scan_get_scanner_promise("fake:scanner0")
        promise = promise.then(
            openpaperwork_core.promise.Promise(
                self.core, Scanner.get_source, args=("fake_source0",)
            )
        )
        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, Source.scan, args=(scan_id,)
            )
        )

        def close(args):
            (source, scan_id, imgs) = args
            source.close()
            return (source, scan_id, imgs)

        promise = promise.then(close)
        return (scan_id, promise)



class AbstractImgEditor(object):
    def transform(self, img, preview=False):
        """
        Apply the transformation operation.

        Arguments:
            img -- image to transform
            preview -- indicates if we intend to use the result as a preview
                or as final result
        """
        assert False

    def transform_frame(self, img_size, frame):
        """
        From the frame applied to the original image to the resulting image.
        """
        return frame

    def untransform_frame(self, img_size, frame):
        """
        From the frame on the resulting image to the original image.
        """
        return frame

    def __eq__(self, o):
        return type(self) is type(o)



import logging

import openpaperwork_core
import openpaperwork_core.promise

from . import (
    DirectFileImporter,
    FileImport,
    RecursiveFileImporter
)
from .. import _


LOGGER = logging.getLogger(__name__)


class SingleImgImporter(object):
    def __init__(self, factory, file_import, src_file_uri):
        self.factory = factory
        self.core = factory.core
        self.file_import = file_import
        self.src_file_uri = src_file_uri
        self.doc_id = None
        self.doc_url = None

    def _append_file_to_doc(self, file_url, doc_id=None):
        if doc_id is None:
            # new document
            (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")
        else:
            # update existing one
            doc_url = self.core.call_success("doc_id_to_url", doc_id)

        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            nb_pages = 0

        # Makes sure it's actually an image and convert it to the expected
        # format
        img = self.core.call_success("url_to_pillow", file_url)
        if img is None:
            LOGGER.error("Failed to load image %s", file_url)
            return (None, None)
        page_url = self.core.call_success(
            "page_get_img_url", doc_url, nb_pages, write=True
        )
        self.core.call_success("pillow_to_url", img, page_url)

        return (doc_id, doc_url)

    def _basic_import(self, file_uri):
        (self.doc_id, self.doc_url) = self._append_file_to_doc(
            file_uri, self.file_import.active_doc_id
        )
        if self.doc_id is None:
            return False

        self.file_import.stats[_("Images")] += 1
        if self.file_import.active_doc_id is None:
            self.file_import.new_doc_ids.add(self.doc_id)
            self.file_import.stats[_("Documents")] += 1
        else:
            self.file_import.upd_doc_ids.add(self.doc_id)
            self.file_import.stats[_("Pages")] += 1

        self.file_import.active_doc_id = self.doc_id

        return True

    def get_promise(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self._basic_import, args=(self.src_file_uri,)
        )


class SingleImgImporterFactory(object):
    def __init__(self, plugin):
        self.plugin = plugin
        self.core = plugin.core

    @staticmethod
    def get_name():
        return _("Append the image to the current document")

    @staticmethod
    def get_recursive_name():
        return _(
            "Find the images recursively and import them to the current"
            " document"
        )

    def is_importable(self, core, file_uri):
        mime = core.call_success("fs_get_mime", file_uri)
        if mime is not None:
            mimes = [mime[1] for mime in Plugin.IMG_MIME_TYPES]
            if mime in mimes:
                return True
            return False
        file_ext = file_uri.split(".")[-1].lower()
        if file_ext in self.plugin.FILE_EXTENSIONS:
            return True

    def get_required_data(self, file_uri):
        return set()

    def make_importer(self, file_import, file_uri, data):
        return SingleImgImporter(self, file_import, file_uri)


class Plugin(openpaperwork_core.PluginBase):
    IMG_MIME_TYPES = [
        ("BMP", "image/x-ms-bmp"),
        ("GIF", "image/gif"),
        ("JPEG", "image/jpeg"),
        ("PNG", "image/png"),
        ("TIFF", "image/tiff"),
    ]

    FILE_EXTENSIONS = [
        "bmp",
        "gif",
        "jpeg",
        "jpg",
        "png",
        "tif",
        "tiff",
    ]

    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return [
            "import"
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'page_img',
                'defaults': ['paperwork_backend.model.img'],
            },
            {
                'interface': 'pillow',
                'defaults': ['openpaperwork_core.pillow.img'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def get_import_mime_types(self, out: set):
        out.update(self.IMG_MIME_TYPES)

    def get_importer(self, out: list, file_import: FileImport):
        importer = DirectFileImporter(
            self.core, file_import, SingleImgImporterFactory(self)
        )
        if importer.can_import():
            out.append(importer)

        importer = RecursiveFileImporter(
            self.core, file_import, SingleImgImporterFactory(self)
        )
        if importer.can_import():
            out.append(importer)



import argparse
import logging
import sys

import openpaperwork_core  # noqa: E402
import openpaperwork_gtk  # noqa: E402

import paperwork_backend  # noqa: E402

# this import must be non-relative due to cx_freeze running this .py
# as an independant Python script
from paperwork_gtk import _  # noqa: E402


LOGGER = logging.getLogger(__name__)

REQUIRED_PLUGINS = (
    'openpaperwork_core.logs.print',
    # plugin 'uncaught_exceptions' requires a mainloop plugin
    'openpaperwork_gtk.mainloop.glib',
)

DEFAULT_GUI_PLUGINS = (
    paperwork_backend.DEFAULT_PLUGINS +
    openpaperwork_gtk.GUI_PLUGINS +
    [
        'openpaperwork_core.display.print',
        'openpaperwork_core.logs.print',
        'openpaperwork_core.spatial.rtree',
        'openpaperwork_gtk.drawer.pillow',
        'openpaperwork_gtk.drawer.scan',
        'openpaperwork_gtk.gesture.autoscrolling',
        'openpaperwork_gtk.gtk_init',
        'paperwork_backend.docscan.autoselect_scanner',
        'paperwork_backend.guesswork.cropping.calibration',
        'paperwork_gtk.about',
        'paperwork_gtk.actions.app.find',
        'paperwork_gtk.actions.app.help',
        'paperwork_gtk.actions.app.open_about',
        'paperwork_gtk.actions.app.open_bug_report',
        'paperwork_gtk.actions.app.open_settings',
        'paperwork_gtk.actions.app.open_shortcuts',
        'paperwork_gtk.actions.doc.add_to_selection',
        'paperwork_gtk.actions.doc.delete',
        'paperwork_gtk.actions.doc.export',
        'paperwork_gtk.actions.doc.new',
        'paperwork_gtk.actions.doc.open_external',
        'paperwork_gtk.actions.doc.prev_next',
        'paperwork_gtk.actions.doc.print',
        'paperwork_gtk.actions.doc.properties',
        'paperwork_gtk.actions.doc.redo_ocr',
        'paperwork_gtk.actions.docs.delete',
        'paperwork_gtk.actions.docs.export',
        'paperwork_gtk.actions.docs.properties',
        'paperwork_gtk.actions.docs.redo_ocr',
        'paperwork_gtk.actions.docs.select_all',
        'paperwork_gtk.actions.page.copy_text',
        'paperwork_gtk.actions.page.delete',
        'paperwork_gtk.actions.page.edit',
        'paperwork_gtk.actions.page.export',
        'paperwork_gtk.actions.page.move_inside_doc',
        'paperwork_gtk.actions.page.move_to_doc',
        'paperwork_gtk.actions.page.print',
        'paperwork_gtk.actions.page.redo_ocr',
        'paperwork_gtk.actions.page.reset',
        'paperwork_gtk.cmd.import',
        'paperwork_gtk.cmd.install',
        'paperwork_gtk.doc_selection',
        'paperwork_gtk.docimport',
        'paperwork_gtk.drawer.calibration',
        'paperwork_gtk.drawer.frame',
        'paperwork_gtk.flatpak',
        'paperwork_gtk.gesture.drag_and_drop',
        'paperwork_gtk.gesture.zoom',
        'paperwork_gtk.icon',
        'paperwork_gtk.keyboard_shortcut.zoom',
        'paperwork_gtk.l10n',
        'paperwork_gtk.mainwindow.doclist',
        'paperwork_gtk.mainwindow.doclist.labeler',
        'paperwork_gtk.mainwindow.doclist.name',
        'paperwork_gtk.mainwindow.doclist.thumbnailer',
        'paperwork_gtk.mainwindow.docproperties',
        'paperwork_gtk.mainwindow.docproperties.extra_text',
        'paperwork_gtk.mainwindow.docproperties.labels',
        'paperwork_gtk.mainwindow.docproperties.name',
        'paperwork_gtk.mainwindow.docview',
        'paperwork_gtk.mainwindow.docview.controllers.autoscrolling',
        'paperwork_gtk.mainwindow.docview.controllers.click',
        'paperwork_gtk.mainwindow.docview.controllers.drop',
        'paperwork_gtk.mainwindow.docview.controllers.empty_doc',
        'paperwork_gtk.mainwindow.docview.controllers.layout',
        'paperwork_gtk.mainwindow.docview.controllers.page_number',
        'paperwork_gtk.mainwindow.docview.controllers.scroll',
        'paperwork_gtk.mainwindow.docview.controllers.title',
        'paperwork_gtk.mainwindow.docview.controllers.zoom',
        'paperwork_gtk.mainwindow.docview.drag',
        'paperwork_gtk.mainwindow.docview.pageadd.buttons',
        'paperwork_gtk.mainwindow.docview.pageadd.import',
        'paperwork_gtk.mainwindow.docview.pageadd.scan',
        'paperwork_gtk.mainwindow.docview.pageadd.source_popover',
        'paperwork_gtk.mainwindow.docview.pageinfo',
        'paperwork_gtk.mainwindow.docview.pageinfo.actions',
        'paperwork_gtk.mainwindow.docview.pageinfo.layout_settings',
        'paperwork_gtk.mainwindow.docview.pageprocessing',
        'paperwork_gtk.mainwindow.docview.pageview',
        'paperwork_gtk.mainwindow.docview.pageview.boxes',
        'paperwork_gtk.mainwindow.docview.pageview.boxes.all',
        'paperwork_gtk.mainwindow.docview.pageview.boxes.hover',
        'paperwork_gtk.mainwindow.docview.pageview.boxes.search',
        'paperwork_gtk.mainwindow.docview.pageview.boxes.selection',
        'paperwork_gtk.mainwindow.docview.progress',
        'paperwork_gtk.mainwindow.docview.scanview',
        'paperwork_gtk.mainwindow.exporter',
        'paperwork_gtk.mainwindow.home',
        'paperwork_gtk.mainwindow.pageeditor',
        'paperwork_gtk.mainwindow.search.advanced',
        'paperwork_gtk.mainwindow.search.field',
        'paperwork_gtk.mainwindow.search.suggestions',
        'paperwork_gtk.mainwindow.window',
        'paperwork_gtk.menus.app.help',
        'paperwork_gtk.menus.app.open_about',
        'paperwork_gtk.menus.app.open_bug_report',
        'paperwork_gtk.menus.app.open_settings',
        'paperwork_gtk.menus.app.open_shortcuts',
        'paperwork_gtk.menus.doc.add_to_selection',
        'paperwork_gtk.menus.doc.delete',
        'paperwork_gtk.menus.doc.export',
        'paperwork_gtk.menus.doc.open_external',
        'paperwork_gtk.menus.doc.print',
        'paperwork_gtk.menus.doc.properties',
        'paperwork_gtk.menus.doc.redo_ocr',
        'paperwork_gtk.menus.docs.delete',
        'paperwork_gtk.menus.docs.export',
        'paperwork_gtk.menus.docs.properties',
        'paperwork_gtk.menus.docs.redo_ocr',
        'paperwork_gtk.menus.docs.select_all',
        'paperwork_gtk.menus.page.copy_text',
        'paperwork_gtk.menus.page.delete',
        'paperwork_gtk.menus.page.export',
        'paperwork_gtk.menus.page.move_inside_doc',
        'paperwork_gtk.menus.page.move_to_doc',
        'paperwork_gtk.menus.page.print',
        'paperwork_gtk.menus.page.redo_ocr',
        'paperwork_gtk.menus.page.reset',
        'paperwork_gtk.model.help',
        'paperwork_gtk.model.help.intro',
        'paperwork_gtk.new_doc',
        'paperwork_gtk.notifications.dialog',
        'paperwork_gtk.notifications.notify',
        'paperwork_gtk.print',
        'paperwork_gtk.settings',
        'paperwork_gtk.settings.ocr.selector_popover',
        'paperwork_gtk.settings.ocr.settings',
        'paperwork_gtk.settings.scanner.calibration',
        'paperwork_gtk.settings.scanner.dev_id_popover',
        'paperwork_gtk.settings.scanner.flatpak',
        'paperwork_gtk.settings.scanner.mode_popover',
        'paperwork_gtk.settings.scanner.resolution_popover',
        'paperwork_gtk.settings.scanner.settings',
        'paperwork_gtk.settings.stats',
        'paperwork_gtk.settings.storage',
        'paperwork_gtk.settings.update',
        'paperwork_gtk.shortcuts.app.find',
        'paperwork_gtk.shortcuts.doc.new',
        'paperwork_gtk.shortcuts.doc.prev_next',
        'paperwork_gtk.shortcuts.doc.print',
        'paperwork_gtk.shortcuts.doc.properties',
        'paperwork_gtk.shortcuts.page.copy_text',
        'paperwork_gtk.shortcuts.page.edit',
        'paperwork_gtk.shortcutswin',
        'paperwork_gtk.sync_on_start',
        'paperwork_gtk.update_notification',
        'paperwork_gtk.widget.flowlayout',
        'paperwork_gtk.widget.label',
    ]
)


class DefaultConsole:
    def print(self, text):
        print(text)

    def input(self, prompt=""):
        return input(prompt)


def gtk_main(app, options, core):
    app = core.call_success("gtk_get_app")
    if app.get_is_remote():
        LOGGER.info("passing control to main paperwork instance")
    # if paperwork is already running, focus it, otherwise remain in this
    # process
    app.activate()
    return 0


def main_main(in_args):
    # To load the plugins, we need first to load the configuration plugin
    # to get the list of plugins to load.
    # The configuration plugin may write traces using logging, so we better
    # enable and configure the plugin logs.print first.

    core = openpaperwork_core.Core()
    for module_name in REQUIRED_PLUGINS:
        core.load(module_name)
    for module_name in paperwork_backend.DEFAULT_CONFIG_PLUGINS:
        core.load(module_name)
    core.init()
    core.call_all(
        "init_logs", "paperwork-gtk",
        "info" if len(in_args) <= 0 else "warning"
    )

    core.call_all("config_load")
    core.call_all("config_load_plugins", "paperwork-gtk", DEFAULT_GUI_PLUGINS)

    if len(in_args) <= 0:
        core.call_all("gtk_init", gtk_main, core)
    else:
        # we remain in this process and call the plugin requested by the
        # command line
        parser = argparse.ArgumentParser()
        cmd_parser = parser.add_subparsers(
            help=_('command'), dest='command', required=True
        )
        core.call_all("cmd_complete_argparse", cmd_parser)
        args = parser.parse_args(in_args)
        console = DefaultConsole()
        core.call_all("cmd_set_console", console)
        core.call_all("cmd_run", console, args)

    return 0


def main():
    main_main(sys.argv[1:])


if __name__ == "__main__":
    # Do not remove. Cx_freeze goes through here
    main()



import collections
import datetime
import locale
import unicodedata

from .. import (_, PluginBase)


class Plugin(PluginBase):
    MONTH_FORMATS = collections.defaultdict(
        lambda: "%B",
        oc="%b",  # Occitan
    )

    def __init__(self):
        self.today = datetime.date.today()
        self.yesterday = self.today - datetime.timedelta(days=1)

        # Need the l10n plugin to be loaded first before getting the
        # translations
        self.i18n_today = None
        self.i18n_yesterday = None
        self.i18n_sizes = ()

    def get_interfaces(self):
        return ['i18n']

    def get_deps(self):
        return [
            {
                'interface': 'l10n',
                'defaults': ['openpaperwork_core.l10n.python'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.i18n_today = _("Today")
        self.i18n_yesterday = _("Yesterday")
        self.i18n_sizes = (
            _('%3.1f bytes'),
            _('%3.1f KiB'),
            _('%3.1f MiB'),
            _('%3.1f GiB'),
            _('%3.1f TiB'),
        )

    def i18n_date_short(self, date):
        if hasattr(date, 'date'):
            date = date.date()  # datetime --> date
        if date == self.today:
            return self.i18n_today
        elif date == self.yesterday:
            return self.i18n_yesterday
        else:
            return date.strftime("%x")

    def i18n_parse_date_short(self, txt):
        if txt == self.i18n_today:
            return self.today
        elif txt == self.i18n_yesterday:
            return self.yesterday
        else:
            try:
                return datetime.datetime.strptime(txt, "%x").date()
            except ValueError:
                return None

    def i18n_date_long_year(self, date):
        if hasattr(date, 'date'):
            date = date.date()  # datetime --> date
        return date.strftime("%Y")

    def i18n_date_long_month(self, date):
        if hasattr(date, 'date'):
            date = date.date()  # datetime --> date
        locale_msg = None
        if hasattr(locale, 'LC_MESSAGES'):
            locale_msg = locale.getlocale(locale.LC_MESSAGES)
        elif hasattr(locale, 'LC_ALL'):
            locale_msg = locale.getlocale(locale.LC_ALL)
        if locale_msg is None or locale_msg[0] is None:
            locale_msg = None
        else:
            locale_msg = locale_msg[0].split("_", 1)[0]
        return date.strftime(self.MONTH_FORMATS[locale_msg])

    def i18n_file_size(self, num):
        for string in self.i18n_sizes:
            if num < 1024.0:
                return string % (num)
            num /= 1024.0
        return self.i18n_sizes[-1] % (num)

    def i18n_strip_accents(self, string):
        """
        Strip all the accents from the string
        """
        return u''.join(
            (
                character for character in unicodedata.normalize('NFD', string)
                if unicodedata.category(character) != 'Mn'
            )
        )

    def i18n_sort(self, string_list):
        t = [
            (self.i18n_strip_accents(str(e).lower()), e)
            for e in string_list
        ]
        t.sort()
        return [e[1] for e in t]



import logging

import openpaperwork_core
import openpaperwork_core.deps

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -400
    MODES = {
        'Color': _("Color"),
        'Gray': _("Grayscale"),
        'Lineart': _("Black & White"),
    }

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.extra_widget = None
        self.config = []

    def get_interfaces(self):
        return [
            'gtk_settings_scanner',
            'screenshot_provider',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'l10n_init',
                'defaults': ['paperwork_gtk.l10n'],
            },
            {
                'interface': 'scan',
                'defaults': ['paperwork_backend.docscan.libinsane'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.config = [
            (
                'settings_scanner_name', 'scanner_device_value',
                _("No scanner selected"), "{}".format
            ),
            (
                'scanner_resolution', 'scanner_resolution_value',
                _("No resolution selected"), _("{} dpi").format
            ),
            (
                'scanner_mode', 'scanner_mode_value',
                _("No mode selected"), self._translate_mode
            ),
        ]

        opt = self.core.call_success(
            "config_build_simple", "settings_scanner", "name",
            lambda: self.core.call_success("config_get", "scanner_dev_id")
        )
        self.core.call_all("config_register", "settings_scanner_name", opt)
        self.core.call_all(
            "config_add_observer", "scanner_dev_id", self._update_scanner_name
        )

    def _update_scanner_name(self):
        def set_scanner_name(devs):
            active = self.core.call_success("config_get", "scanner_dev_id")
            for dev in devs:
                dev_id = dev[0]
                dev_name = dev[1]
                if dev_id == active:
                    self.core.call_success(
                        "config_put", "settings_scanner_name", dev_name
                    )
                    break
            return devs

        promise = self.core.call_success("scan_list_scanners_promise")
        promise = promise.then(set_scanner_name)
        self.core.call_success("scan_schedule", promise)

    def complete_settings(self, global_widget_tree):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings.scanner",
            "settings.glade"
        )
        self.widget_tree = widget_tree

        extra_widget = self.core.call_success(
            "settings_scanner_get_extra_widget"
        )

        self.core.call_success(
            "add_setting_to_dialog", global_widget_tree,
            _("Scanner"),
            [
                widget_tree.get_object("scanner_device"),
                widget_tree.get_object("scanner_resolution"),
                widget_tree.get_object("scanner_mode"),
                widget_tree.get_object("scanner_calibration"),
            ],
            extra_widget=extra_widget
        )

        def refresh(*args, **kwargs):
            self._refresh_settings(widget_tree)

        def disable_refresh(*args, **kwargs):
            for c in self.config:
                self.core.call_all("config_remove_observer", c[0], refresh)

        for c in self.config:
            self.core.call_all("config_add_observer", c[0], refresh)

        global_widget_tree.get_object("settings_window").connect(
            "destroy", disable_refresh
        )

        self._refresh_settings(widget_tree)

        list_scanners_promise = self.core.call_success(
            "scan_list_scanners_promise"
        )
        self.core.call_all(
            "complete_scanner_settings",
            global_widget_tree, widget_tree,
            list_scanners_promise
        )
        self.core.call_success("scan_schedule", list_scanners_promise)

    def _translate_mode(self, mode):
        if mode in self.MODES:
            return self.MODES[mode]
        return mode

    def _refresh_settings(self, widget_tree):
        for (config_key, widget_name, default_value, fmt) in self.config:
            value = self.core.call_success("config_get", config_key)
            if value is not None:
                value = fmt(value)
            else:
                value = default_value
            widget_tree.get_object(widget_name).set_text(value)

        active = self.core.call_success("config_get", "scanner_dev_id")
        active = active is not None and active != ""
        buttons = [
            'scanner_resolution',
            'scanner_mode',
            'scanner_calibration',
        ]
        for button in buttons:
            # WORKAROUND(Jflesch): set_sensitive() doesn't appear to work on
            # GtkMenuButton
            widget_tree.get_object(button).set_sensitive(active)

    def screenshot_snap_all_doc_widgets(self, out_dir):
        if self.widget_tree is None:
            return

        buttons = [
            "scanner_device",
            "scanner_resolution",
            "scanner_mode",
            "scanner_calibration",
        ]
        for button_name in buttons:
            button = self.widget_tree.get_object(button_name)
            self.core.call_success(
                "screenshot_snap_widget", button,
                self.core.call_success(
                    "fs_join", out_dir, "settings_{}.png".format(button_name)
                ),
                margins=(100, 100, 100, 100)
            )



import logging

import PIL
import PIL.Image

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_core.promise

from . import (
    AbstractExportPipe,
    AbstractExportPipePlugin,
    ExportDataType
)
from .. import _


CAIRO_AVAILABLE = False
GI_AVAILABLE = False
GLIB_AVAILABLE = False
PANGO_AVAILABLE = False

try:
    import cairo
    CAIRO_AVAILABLE = True
except (ImportError, ValueError):
    pass

try:
    import gi
    GI_AVAILABLE = True
except (ImportError, ValueError):
    pass

if GI_AVAILABLE:
    try:
        gi.require_version('Pango', '1.0')
        gi.require_version('PangoCairo', '1.0')
        from gi.repository import Pango
        from gi.repository import PangoCairo
        PANGO_AVAILABLE = True
    except (ImportError, ValueError):
        pass


LOGGER = logging.getLogger(__name__)


class PdfDocUrlToPdfUrlExportPipe(AbstractExportPipe):
    """
    Simply copy the PDF we have.
    """
    def __init__(self, core):
        super().__init__(
            name="unmodified_pdf",
            input_type=ExportDataType.DOCUMENT,
            output_type=ExportDataType.OUTPUT_URL_FILE
        )
        self.core = core

    def can_export_doc(self, doc_url):
        pdf_url = self.core.call_success("doc_get_pdf_url_by_url", doc_url)
        return pdf_url is not None

    def export(self, input_data, result='final', target_file_url=None):
        if target_file_url is None:
            (target_file_url, file_desc) = self.core.call_success(
                "fs_mktemp", prefix="paperwork-export-", suffix=".pdf",
                mode="wb", on_disk=True
            )
            file_desc.close()

        list_docs = input_data.iter(ExportDataType.DOCUMENT)
        out_files = []
        for (idx, (doc_set, doc)) in enumerate(list_docs):
            out = target_file_url
            if idx != 0:
                out = target_file_url.rsplit(".", 1)
                out = "{}_{}.{}".format(out[0], idx, out[1])

            assert not doc.expanded
            pdf_url = self.core.call_success(
                "doc_get_pdf_url_by_url", doc.data[1]
            )
            assert pdf_url is not None
            self.core.call_success("fs_copy", pdf_url, out)
            out_files.append(out)
        return out_files

    def get_output_mime(self):
        return ("application/pdf", ("pdf",))

    def __str__(self):
        return _("Original PDF(s)")


class PdfCreator(object):
    def __init__(self, core, target_file_url, page_format, quality):
        self.core = core
        self.page_format = page_format
        self.quality = quality

        self.file_descriptor = core.call_success(
            "fs_open", target_file_url, 'wb'
        )

        self.pdf_surface = cairo.PDFSurface(
            self.file_descriptor, self.page_format[0], self.page_format[1]
        )
        self.pdf_context = cairo.Context(self.pdf_surface)

        self.gc_protection = []  # WORKAROUND(Jflesch): Cairo crash

    def set_page_size(self, img_size):
        # portrait or landscape
        if (img_size[0] < img_size[1]):
            self.pdf_size = (
                min(self.page_format[0], self.page_format[1]),
                max(self.page_format[0], self.page_format[1])
            )
        else:
            self.pdf_size = (
                max(self.page_format[0], self.page_format[1]),
                min(self.page_format[0], self.page_format[1])
            )
        self.pdf_surface.set_size(self.pdf_size[0], self.pdf_size[1])

    def paint_txt(self, boxes, img_size):
        scale_factor_x = self.pdf_size[0] / img_size[0]
        scale_factor_y = self.pdf_size[1] / img_size[1]
        scale_factor = min(scale_factor_x, scale_factor_y)

        for line in boxes:
            for word in line.word_boxes:
                if word.position[0][0] <= 0 and line.position[0][1] <= 0:
                    continue

                box_size = (
                    (word.position[1][0] - word.position[0][0])
                    * scale_factor,
                    (line.position[1][1] - line.position[0][1])
                    * scale_factor
                )
                if 0 in box_size:
                    continue

                layout = PangoCairo.create_layout(self.pdf_context)
                layout.set_text(word.content, -1)

                txt_size = layout.get_size()
                if 0 in txt_size:
                    continue

                txt_factors = (
                    float(box_size[0]) * Pango.SCALE / txt_size[0],
                    float(box_size[1]) * Pango.SCALE / txt_size[1],
                )

                self.pdf_context.save()
                try:
                    self.pdf_context.set_source_rgb(0, 0, 0)
                    self.pdf_context.translate(
                        word.position[0][0] * scale_factor,
                        line.position[0][1] * scale_factor
                    )

                    # make the text use the whole box space
                    self.pdf_context.scale(txt_factors[0], txt_factors[1])

                    PangoCairo.update_layout(self.pdf_context, layout)
                    PangoCairo.show_layout(self.pdf_context, layout)
                finally:
                    self.pdf_context.restore()
        return self

    def paint_img(self, img):
        img_size = img.size
        scale_factor_x = self.pdf_size[0] / img_size[0]
        scale_factor_y = self.pdf_size[1] / img_size[1]
        scale_factor = min(scale_factor_x, scale_factor_y)

        img_surface = self.core.call_success(
            "pillow_to_surface", img,
            intermediate="jpeg", quality=int(self.quality * 100)
        )
        # WORKAROUND(Jflesch):
        # If Cairo supports JPEG embedding, we use
        # cairo.ImageSurface.set_mime_data() instead of the usual Cairo
        # surface functions to draw the image. It seems this function does
        # not increment the ref counter of the surface object
        # --> the Python GC tends to collect it while Cairo is still going
        # to use it to generate the PDF.
        # --> we have to keep a reference on it ourselves, until the page has
        # been generated.
        self.gc_protection.append(img_surface)

        self.pdf_context.save()
        try:
            self.pdf_context.identity_matrix()
            self.pdf_context.scale(scale_factor, scale_factor)
            self.pdf_context.set_source_surface(img_surface.surface)
            self.pdf_context.paint()
        finally:
            self.pdf_context.restore()

    def next_page(self):
        self.pdf_context.show_page()
        self.gc_protection = []  # WORKAROUND(Jflesch): Cairo crash

    def finish(self):
        self.pdf_surface.flush()
        self.pdf_surface.finish()
        self.file_descriptor.close()


class PagesToPdfUrlExportPipe(AbstractExportPipe):
    def __init__(self, core):
        super().__init__(
            name="generated_pdf",
            input_type=ExportDataType.IMG_BOXES,
            output_type=ExportDataType.OUTPUT_URL_FILE
        )
        self.core = core
        self.can_change_quality = True
        self.can_change_page_format = True

    def set_page_format(self, page_format):
        self.page_format = page_format

    def export(self, input_data, result='final', target_file_url=None):
        if target_file_url is None:
            (target_file_url, file_desc) = self.core.call_success(
                "fs_mktemp", prefix="paperwork-export-", suffix=".pdf",
                mode="w", on_disk=True
            )
            # we need the file name, not the file descriptor
            file_desc.close()

        last_doc = None
        doc_idx = 0
        out = target_file_url
        list_img_boxes = input_data.iter(ExportDataType.IMG_BOXES)

        creator = self.core.call_one(
            "mainloop_execute", PdfCreator,
            self.core, target_file_url, self.page_format, self.quality
        )

        out_files = []

        for (doc_set, (doc, (page, img_boxes))) in list_img_boxes:

            if doc.data[1] != last_doc and last_doc is not None:
                # another doc, another output file, another PDFCreator
                self.core.call_one("mainloop_execute", creator.finish)
                out_files.append(out)
                doc_idx += 1
                out = target_file_url.rsplit(".", 1)
                out = "{}_{}.{}".format(out[0], doc_idx, out[1])
                creator = self.core.call_one(
                    "mainloop_execute", PdfCreator,
                    self.core, out, self.page_format, self.quality
                )
            last_doc = doc.data[1]

            (img, boxes) = img_boxes.data
            self.core.call_one(
                "mainloop_execute", creator.set_page_size, img.size
            )
            self.core.call_one(
                "mainloop_execute", creator.paint_txt, boxes,
                img.size
            )
            img_size = img.size
            img_size = (
                int(self.quality * img_size[0]),
                int(self.quality * img_size[1])
            )
            img = img.resize(
                img_size,
                getattr(PIL.Image, 'Resampling', PIL.Image).LANCZOS
            )
            self.core.call_one("mainloop_execute", creator.paint_img, img)
            self.core.call_one("mainloop_execute", creator.next_page)

        self.core.call_one("mainloop_execute", creator.finish)
        out_files.append(out)
        return out_files

    def get_output_mime(self):
        return ("application/pdf", ("pdf",))

    def __str__(self):
        return _("Generated PDF(s)")


class Plugin(AbstractExportPipePlugin):
    def init(self, core):
        super().init(core)
        self.pipes = [
            PdfDocUrlToPdfUrlExportPipe(core),
            PagesToPdfUrlExportPipe(core),
        ]

    def get_interfaces(self):
        return super().get_interfaces() + [
            "chkdeps",
            "export_pipes_pdf",
        ]

    def get_deps(self):
        return super().get_deps() + [
            {
                'interface': 'pillow_to_surface',
                'defaults': ['paperwork_backend.cairo.pillow'],
            },
        ]

    def chkdeps(self, out: dict):
        if not CAIRO_AVAILABLE:
            out['cairo'].update(openpaperwork_core.deps.CAIRO)
        if not GI_AVAILABLE:
            out['gi'].update(openpaperwork_core.deps.GI)
        if not PANGO_AVAILABLE:
            out['pango'].update(openpaperwork_core.deps.PANGO)



import logging

import PIL

import openpaperwork_core

from paperwork_backend.model.thumbnail import (
    THUMBNAIL_HEIGHT,
    THUMBNAIL_WIDTH
)

from ... import _


LOGGER = logging.getLogger(__name__)
DELAY = 0.01

# placeholders size must include borders
PLACEHOLDER_HEIGHT = THUMBNAIL_HEIGHT + 2
PLACEHOLDER_WIDTH = THUMBNAIL_WIDTH + 2


class ThumbnailTask(object):
    def __init__(self, plugin, doc_id, gtk_image):
        self.plugin = plugin
        self.core = plugin.core

        self.doc_id = doc_id
        self.gtk_image = gtk_image

    def set_thumbnail(self, img=None):
        if img is None:
            LOGGER.warning(
                "Failed to get thumbnail for document %s", self.doc_id
            )
            return
        pixbuf = self.core.call_success("pillow_to_pixbuf", img)
        self.gtk_image.set_from_pixbuf(pixbuf)

    def get_promise(self):
        doc_url = self.core.call_success("doc_id_to_url", self.doc_id)
        if doc_url is None:
            return openpaperwork_core.promise.Promise(self.core)

        def on_error(exc):
            # catch any error, otherwise if a document is corrupted,
            # the user will get far too many error messages during
            # thumbnailing
            LOGGER.error(
                "Failed to generate thumbnail of document %s",
                self.doc_id, exc_info=exc
            )

        promise = openpaperwork_core.promise.Promise(
            self.core,
            LOGGER.info, args=("Thumbnailing of document %s", self.doc_id,)
        )
        promise = promise.then(lambda *args: None)  # drop logger return value
        promise = promise.then(
            self.core.call_success("thumbnail_get_doc_promise", doc_url)
        )
        promise = promise.then(self.set_thumbnail)
        promise = promise.catch(on_error)
        return promise


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def __init__(self):
        super().__init__()
        self.default_thumbnail = None
        self.running = False

        self.nb_loaded = 0
        self.nb_to_load = 0
        self._progress_str = None

        self.processing_docs = set()
        self.processed_docs = set()

    def get_interfaces(self):
        return [
            'gtk_doclist_listener',
            'gtk_thumbnailer',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.maindow.doclist'],
            },
            {
                'interface': 'pillow_util',
                'defaults': ['openpaperwork_core.pillow.util'],
            },
            {
                'interface': 'pixbuf_pillow',
                'defaults': ['openpaperwork_gtk.pixbuf.pillow'],
            },
            {
                'interface': 'thumbnail',
                'defaults': ['paperwork_backend.model.thumbnail'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def init(self, core):
        super().init(core)
        img = PIL.Image.new(
            "RGB",
            (THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT),
            color="#EEEEEE"
        )
        self.default_thumbnail = self.core.call_success(
            "pillow_to_pixbuf", img
        )
        self.core.call_all(
            "work_queue_create", "thumbnailer", stop_on_quit=True
        )
        self._progress_str = _("Loading document thumbnails")

    def doclist_show(self, docs):
        self.core.call_all("work_queue_cancel_all", "thumbnailer")

    def on_doc_list_clear(self):
        self.processed_docs = set()
        self.on_doc_list_visibility_changed()

    def on_doc_list_visibility_changed(self):
        self.core.call_all("work_queue_cancel_all", "thumbnailer")
        self.processing_docs = set()
        self.nb_to_load = 0

    def on_doc_box_creation(self, doc_id, gtk_row, gtk_custom_flowlayout):
        gtk_img = gtk_row.thumbnail
        gtk_img.set_size_request(PLACEHOLDER_WIDTH, PLACEHOLDER_HEIGHT)
        gtk_img.set_visible(True)

    def on_doc_box_visible(self, doc_id, gtk_row, gtk_custom_flowlayout):
        if doc_id in self.processing_docs:
            return
        if doc_id in self.processed_docs:
            return

        self.nb_to_load += 1
        self.processing_docs.add(doc_id)

        task = ThumbnailTask(self, doc_id, gtk_row.thumbnail)
        # Gives back a bit of CPU time to GTK so the GUI remains
        # usable
        promise = openpaperwork_core.promise.DelayPromise(
            self.core, DELAY
        )
        promise = promise.then(task.get_promise())

        def _when_loaded():
            self.nb_loaded += 1
            if doc_id in self.processing_docs:
                self.processing_docs.remove(doc_id)
                self.processed_docs.add(doc_id)
            self._update_progress()

        promise = promise.then(_when_loaded)

        self.core.call_success(
            "work_queue_add_promise", "thumbnailer", promise
        )

    def _update_progress(self):
        if self.nb_loaded > self.nb_to_load:
            self.nb_loaded = 0
            self.nb_to_load = 0
            self.core.call_all("on_progress", "thumbnailing", 1.0)
            return

        self.core.call_all(
            "on_progress", "thumbnailing", self.nb_loaded / self.nb_to_load,
            self._progress_str
        )



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    # so we can report an accurate document count
    PRIORITY = 1000

    def __init__(self):
        super().__init__()
        self.selection = set()

    def get_interfaces(self):
        return ['doc_selection']

    def doc_selection_reset(self):
        self.selection = set()

    def doc_selection_add(self, doc_id, doc_url):
        self.selection.add((doc_id, doc_url))

    def doc_selection_remove(self, doc_id, doc_url):
        try:
            self.selection.remove((doc_id, doc_url))
        except KeyError:
            pass

    def doc_selection_get(self, out: set):
        out.update(self.selection)

    def doc_selection_len(self):
        nb_docs = len(self.selection)
        if nb_docs <= 0:
            return None
        return nb_docs

    def doc_selection_in(self, doc_id, doc_url):
        if (doc_id, doc_url) in self.selection:
            return True
        return None



import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['nb_pages']

    def doc_get_nb_pages_by_url(self, doc_url):
        out = []
        self.core.call_all("doc_internal_get_nb_pages_by_url", out, doc_url)
        r = max(out, default=0)
        if r == 0:
            return None
        return r

    def page_get_hash_by_url(self, doc_url, page_idx):
        out = []
        self.core.call_all(
            "page_internal_get_hash_by_url", out, doc_url, page_idx
        )
        r = 0
        for h in out:
            r ^= h
        return r

    def doc_get_mtime_by_url(self, doc_url):
        out = []
        self.core.call_all("doc_internal_get_mtime_by_url", out, doc_url)
        return max(out, default=None)

    def page_get_mtime_by_url(self, doc_url, page_idx):
        out = []
        self.core.call_all(
            "page_internal_get_mtime_by_url", out, doc_url, page_idx
        )
        return max(out, default=None)



import pillowfight

import openpaperwork_core

from . import AbstractImgEditor


class ColorImgEditor(AbstractImgEditor):
    def transform(self, img, preview=False):
        return pillowfight.ace(
            img, samples=50 if preview else 200
        )


class Plugin(openpaperwork_core.PluginBase):
    NAME = 'color_equalization'

    def get_interfaces(self):
        return ['img_editor']

    def img_editor_get_names(self, out: list):
        out.append(self.NAME)

    def img_editor_get(self, name, *args, **kwargs):
        if name != self.NAME:
            return None
        return ColorImgEditor()

    def img_editor_set(self, inout: list, name, *args, **kwargs):
        if name != self.NAME:
            return None
        c = ColorImgEditor()
        if c in inout:
            # an ACE editor is already in the list
            return
        inout.append(c)

    def img_editor_unset(self, inout: list, name):
        if name != self.NAME:
            return None
        c = ColorImgEditor()
        while c in inout:
            inout.remove(c)



import logging

try:
    from gi.repository import Gio
    GIO_AVAILABLE = True
except (ImportError, ValueError):
    GIO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'action',
            'action_app',
            'action_app_help',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'help_documents',
                'defaults': ['paperwork_gtk.model.help'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GIO_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def init(self, core):
        super().init(core)
        if not GIO_AVAILABLE:
            return
        for (title, file_name) in self.core.call_success("help_get_files"):
            action = Gio.SimpleAction.new('open_help.' + file_name, None)
            action.connect("activate", self.doc_open_help, file_name)
            self.core.call_all("app_actions_add", action)

    def doc_open_help(self, action, parameter, file_name):
        doc_url = self.core.call_success("help_get_file", file_name)
        self.core.call_all("doc_open", "help_" + file_name, doc_url)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_export_many"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.active_doc = (None, None)
        self.active_page_idx = 0

    def get_interfaces(self):
        return [
            'action',
            'action_docs',
            'action_docs_export',
            'chkdeps',
            'doc_open',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_selection',
                'defaults': ['paperwork_gtk.doc_selection'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._export)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)

    def on_page_shown(self, active_page_idx):
        self.active_page_idx = active_page_idx

    def _export(self, action, parameter):
        docs = set()
        self.core.call_all("doc_selection_get", docs)
        self.core.call_all(
            "gtk_open_exporter_multiple_docs", docs,
            self.active_doc[0], self.active_doc[1], self.active_page_idx
        )



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "page_copy_text"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.selected_text = ""
        self.windows = []

    def get_interfaces(self):
        return [
            'action',
            'action_page',
            'action_page_copy_text',
            'chkdeps',
            'selected_boxes_listener',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return

        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._copy_text)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def on_page_boxes_selected(self, doc_id, doc_url, page_id, boxes):
        self.selected_text = " ".join([b.content for b in boxes])

    def _copy_text(self, *args, **kwargs):
        LOGGER.info(
            "Copying %d characters to clipboard", len(self.selected_text)
        )
        clipboard = Gtk.Clipboard.get_default(self.windows[-1].get_display())
        clipboard.set_text(self.selected_text, -1)
